import logging
from enum import Enum
from typing import Optional, Dict, Set, Any
from whendo.core.util import Now, Http, KeyTagMode, DateTime2, DateTime
from whendo.core.hooks import DispatcherHooks
from whendo.core.action import Action, ActionRez, Rez
from whendo.core.resolver import resolve_rez


logger = logging.getLogger(__name__)


class DispatcherAction(Action):
    """
    Instances of subclasses of DispatcherAction invoke Dispatcher
    methods directly by way of the class, DispatcherHooks.
    """


class ScheduleProgram(DispatcherAction):
    program_name: Optional[str] = None
    start_stop: Optional[DateTime2] = None
    schedule_program: str = "schedule_program"

    def description(self):
        return f"This action schedules a program ({self.program_name}) start_stop ({self.start_stop})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        program_name = flds.get("program_name", None)
        if program_name == None:
            raise ValueError("program name missing")
        start_stop = flds.get("start_stop", None)
        if start_stop == None:
            raise ValueError(f"start_stop missing")
        DispatcherHooks.schedule_program(
            program_name=program_name, start=start_stop.dt1, stop=start_stop.dt2
        )
        result = f"program ({program_name}) scheduled, start_stop({start_stop})"
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class UnscheduleProgram(DispatcherAction):
    program_name: Optional[str] = None
    unschedule_program: str = "unschedule_program"

    def description(self):
        return f"This action unschedules a program ({self.program_name})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        program_name = flds.get("program_name", None)
        if program_name == None:
            raise ValueError("program name missing")
        DispatcherHooks.unschedule_program(program_name=program_name)
        result = f"program ({program_name}) unscheduled"
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class ScheduleAction(DispatcherAction):
    scheduler_name: Optional[str] = None
    action_name: Optional[str] = None
    schedule_action: str = "schedule_action"

    def description(self):
        return f"This action schedules an action ({self.action_name}) using scheduler ({self.scheduler_name})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        scheduler_name = flds.get("scheduler_name", None)
        if scheduler_name == None:
            raise ValueError("scheduler name missing")
        action_name = flds.get("action_name", None)
        if action_name == None:
            raise ValueError("action name missing")
        DispatcherHooks.schedule_action(
            scheduler_name=scheduler_name, action_name=action_name
        )
        result = f"action ({action_name}) scheduled using scheduler ({scheduler_name})"
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class UnscheduleSchedulerAction(DispatcherAction):
    scheduler_name: Optional[str] = None
    action_name: Optional[str] = None
    unschedule_scheduler_action: str = "unschedule_scheduler_action"

    def description(self):
        return f"This action unschedules an action ({self.action_name}) with scheduler ({self.scheduler_name})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        scheduler_name = flds.get("scheduler_name", None)
        if scheduler_name == None:
            raise ValueError("scheduler name missing")
        action_name = flds.get("action_name", None)
        if action_name == None:
            raise ValueError("action name missing")
        DispatcherHooks.unschedule_scheduler_action(
            scheduler_name=scheduler_name, action_name=action_name
        )
        result = f"action ({action_name}) unscheduled from scheduler ({scheduler_name})"
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class UnscheduleScheduler(DispatcherAction):
    scheduler_name: Optional[str] = None
    unschedule_scheduler: str = "unschedule_scheduler"

    def description(self):
        return f"This action unschedules a scheduler ({self.scheduler_name})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        scheduler_name = flds.get("scheduler_name", None)
        if scheduler_name == None:
            raise ValueError("scheduler name missing")
        DispatcherHooks.unschedule_scheduler(scheduler_name=scheduler_name)
        result = f"scheduler ({scheduler_name}) unscheduled"
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class UnscheduleAllSchedulers(DispatcherAction):
    unschedule_all_schedulers: str = "unschedule_all_schedulers"

    def description(self):
        return f"This action unschedules all schedulers."

    def execute(self, tag: str = None, rez: Rez = None):
        DispatcherHooks.unschedule_all_schedulers()
        result = f"all schedulers unscheduled"
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class DeferAction(DispatcherAction):
    scheduler_name: Optional[str] = None
    action_name: Optional[str] = None
    wait_until: Optional[DateTime] = None
    defer_action: str = "defer_action"

    def description(self):
        return f"This action defers a schedule/action with mode ({self.mode}) and fields: scheduler_name ({self.scheduler_name}), action_name ({self.action_name}), wait_until ({self.wait_until})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        scheduler_name = flds.get("scheduler_name", None)
        if scheduler_name == None:
            raise ValueError("scheduler name missing")
        action_name = flds.get("action_name", None)
        if not action_name:
            raise ValueError("action name missing")
        wait_until = flds.get("wait_until", None)
        if not wait_until:
            raise ValueError("wait_until missing")
        DispatcherHooks.defer_action(
            scheduler_name=scheduler_name,
            action_name=action_name,
            wait_until=wait_until.dt,
        )
        result = f"action ({action_name}) using scheduler ({scheduler_name}) deferred until ({wait_until})"

        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class ExpireAction(DispatcherAction):
    scheduler_name: Optional[str] = None
    action_name: Optional[str] = None
    expire_on: Optional[DateTime] = None
    expire_action: str = "expire_action"

    def description(self):
        return f"This action expires a schedule/action with mode ({self.mode}) and fields: scheduler_name ({self.scheduler_name}), action_name ({self.action_name}), expire_on ({self.expire_on})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        scheduler_name = flds.get("scheduler_name", None)
        if scheduler_name == None:
            raise ValueError("scheduler name missing")
        action_name = flds.get("action_name", None)
        if action_name == None:
            raise ValueError("action name missing")
        expire_on = flds.get("expire_on", None)
        if expire_on == None:
            raise ValueError("expire_on missing")
        DispatcherHooks.expire_action(
            scheduler_name=scheduler_name,
            action_name=action_name,
            expire_on=expire_on.dt,
        )
        result = f"action ({action_name}) using scheduler ({scheduler_name}) expiring on ({expire_on})"
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class ClearAllDeferredActions(DispatcherAction):
    clear_all_deferred_actions: str = "clear_all_deferred_actions"

    def description(self):
        return "Removes all deferred scheduled actions."

    def execute(self, tag: str = None, rez: Rez = None):
        DispatcherHooks.clear_all_deferred_actions()
        result = "All deferred scheduled actions removed."
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class ClearAllExpiringActions(DispatcherAction):
    clear_all_expiring_actions: str = "clear_all_expiring_actions"

    def description(self):
        return "Removes all expiring scheduled actions."

    def execute(self, tag: str = None, rez: Rez = None):
        DispatcherHooks.clear_all_expiring_actions()
        result = "All expiring scheduled actions removed."
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class ClearAllScheduling(DispatcherAction):
    clear_all_scheduling: str = "clear_all_scheduling"

    def description(self):
        return "Removes all scheduled actions, current or planned, and foreground Timed instance jobs. Ignores the 'inventory' objects and the out-of-band Timed instance."

    def execute(self, tag: str = None, rez: Rez = None):
        DispatcherHooks.clear_all_scheduling()
        result = "All scheduling artifacts cleared or reset."
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class Exec(DispatcherAction):
    """
    Execute an action at a server.
    """

    server_name: Optional[str] = None
    action_name: Optional[str] = None
    exec: str = "exec"

    def description(self):
        return f"This action executes ({self.action_name}) at the server ({self.server_name})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        server_name = flds.get("server_name", None)
        action_name = flds.get("action_name", None)
        if action_name == None:
            raise ValueError(f"action name missing")
        if server_name == None:
            host = self.local_host()
            port = self.local_port()
            is_local = True
        else:
            server = DispatcherHooks.get_server(server_name)
            host = server.host
            port = server.port
            is_local = False

        if rez:
            if is_local:
                # execute locally
                result = DispatcherHooks.get_action(action_name).execute(
                    tag=tag, rez=rez
                )
            else:
                response = Http(host=host, port=port).post(
                    f"/actions/{action_name}/execute", rez
                )
                result = resolve_rez(response)
        else:
            if is_local:
                # execute locally
                result = DispatcherHooks.get_action(action_name).execute(tag=tag)
            else:
                response = Http(host=server.host, port=server.port).get(
                    f"/actions/{action_name}/execute"
                )
                result = resolve_rez(response)

        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class ExecKeyTags(DispatcherAction):
    """
    Execute an action at zero or more servers. If key_tags is not provided, executes action at all servers.
    """

    action_name: Optional[str] = None
    key_tags: Optional[Dict[str, Set[str]]] = None
    key_tag_mode: Optional[KeyTagMode] = None
    exec_key_tags: str = "exec_key_tags"

    def description(self):
        return f"This action executes ({self.action_name}) at the servers with key:tags satisfying ({self.key_tags}) using key tag mode ({self.key_tag_mode})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        action_name = flds.get("action_name", None)
        if action_name == None:
            raise ValueError(f"action name missing")
        key_tags = flds.get("key_tags", None)
        key_tag_mode = flds.get("key_tag_mode", KeyTagMode.ANY)

        if key_tags:
            servers = DispatcherHooks.get_servers_by_tags(
                key_tags=key_tags, key_tag_mode=key_tag_mode
            )
        else:
            servers = DispatcherHooks.get_servers()
        result = []
        if rez:
            for server in servers:
                if (
                    server.host == self.local_host()
                    and server.port == self.local_port()
                ):
                    # execute locally
                    result.append(
                        DispatcherHooks.get_action(action_name).execute(
                            tag=tag, rez=rez
                        )
                    )
                else:
                    response = Http(host=server.host, port=server.port).post(
                        f"/actions/{action_name}/execute", rez
                    )
                    result.append(resolve_rez(response))
        else:
            for server in servers:
                if (
                    server.host == self.local_host()
                    and server.port == self.local_port()
                ):
                    # execute locally
                    result.append(
                        DispatcherHooks.get_action(action_name).execute(tag=tag)
                    )
                else:
                    response = Http(host=server.host, port=server.port).get(
                        f"/actions/{action_name}/execute"
                    )
                    result.append(resolve_rez(response))
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class ExecSupplied(DispatcherAction):
    """
    Execute an action at a server.
    """

    server_name: Optional[str] = None
    action: Optional[Action] = None
    exec_supplied: str = "exec_supplied"

    def description(self):
        return (
            f"This action executes ({self.action}) at the server ({self.server_name})."
        )

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        action = flds.get("action", None)
        if action == None:
            raise ValueError(f"action missing")
        server_name = flds.get("server_name", None)
        if server_name == None:
            host = self.local_host()
            port = self.local_port()
            is_local = True
        else:
            server = DispatcherHooks.get_server(server_name)
            host = server.host
            port = server.port
            is_local = False
        print("REZ", rez.result if rez else None)
        print("ACTION", action)
        if rez:
            if is_local:
                # execute locally
                result = action.execute(tag=tag, rez=rez)
            else:
                action_rez = ActionRez(action=action, rez=rez)
                response = Http(host=host, port=port).post(
                    f"/execution/with_rez", action_rez
                )
                result = resolve_rez(response)
        else:
            if is_local:
                # execute locally
                result = action.execute(tag=tag)
            else:
                response = Http(host=server.host, port=server.port).post(
                    f"/execution", action
                )
                result = resolve_rez(response)
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})


class ExecSuppliedKeyTags(DispatcherAction):
    """
    Execute an action at zero or more servers. If key_tags is not provided, executes action at all servers.
    """

    action: Optional[Action] = None
    key_tags: Optional[Dict[str, Set[str]]] = None
    key_tag_mode: Optional[KeyTagMode] = None
    exec_supplied_key_tags: str = "exec_supplied_key_tags"

    def description(self):
        return f"This action executes ({self.action}) at the servers with key:tags satisfying ({self.key_tags}) using key tag mode ({self.key_tag_mode})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        action = flds.get("action", None)
        if action == None:
            raise ValueError(f"action missing")
        key_tags = flds.get("key_tags", None)
        key_tag_mode = flds.get("key_tag_mode", KeyTagMode.ANY)
        if key_tags:
            servers = DispatcherHooks.get_servers_by_tags(
                key_tags=key_tags, key_tag_mode=key_tag_mode
            )
        else:
            servers = DispatcherHooks.get_servers()
        result = []
        if rez:
            for server in servers:
                if (
                    server.host == self.local_host()
                    and server.port == self.local_port()
                ):
                    # execute locally
                    result.append(action.execute(tag=tag, rez=rez))
                else:
                    action_rez = ActionRez(action=action, rez=rez)
                    response = Http(host=server.host, port=server.port).post(
                        f"/execution/with_rez", action_rez
                    )
                    result.append(resolve_rez(response))
        else:
            for server in servers:
                if (
                    server.host == self.local_host()
                    and server.port == self.local_port()
                ):
                    # execute locally
                    result.append(action.execute(tag=tag))
                else:
                    response = Http(host=server.host, port=server.port).post(
                        f"/execution", action
                    )
                    result.append(resolve_rez(response))
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})



from pydantic import BaseModel
from typing import Dict, List, Set, Callable
from datetime import datetime
from collections import namedtuple
import logging
from .util import Now, str_to_dt, dt_to_str


logger = logging.getLogger(__name__)


"""
Instances of the namedtuple, SchedulerActions, provide the fundamental mappings of schedulers to actions.

These objects also form the basis for the extended properties of schedulers
such as the deferral of starting a scheduler and the expiration of a running
scheduler. See the Dispatcher class for usage.
"""
SchedulerActions = namedtuple("SchedulerActions", ["scheduler_name", "action_names"])

"""
A ScheduledActions instance represents a set of active schedulers and their associated
actions. If you want to know which schedulers and actions are active, this is the object
tells the story.
"""


class ScheduledActions(BaseModel):
    scheduler_actions: List[SchedulerActions] = []

    def add(self, scheduler_name: str, action_name: str):
        for sa in self.scheduler_actions:
            if sa.scheduler_name == scheduler_name:
                sa.action_names.add(action_name)
                return
        new_scheduler_action = SchedulerActions(
            scheduler_name=scheduler_name, action_names={action_name}
        )
        self.scheduler_actions.append(new_scheduler_action)

    def actions(self, scheduler_name: str):
        for sa in self.scheduler_actions:
            if sa.scheduler_name == scheduler_name:
                return sa.action_names
        return set()

    def delete(self, scheduler_name: str, action_name: str):
        to_remove = None
        for sa in self.scheduler_actions:
            if sa.scheduler_name == scheduler_name:
                if action_name in sa.action_names:
                    sa.action_names.remove(action_name)
                    if len(sa.action_names) == 0:
                        to_remove = sa
                break
        if to_remove:
            self.scheduler_actions.remove(to_remove)
            return to_remove.scheduler_name
        else:
            return None

    def delete_action(self, action_name: str):
        removed_scheduler_names: Set[str] = set()
        for scheduler_name in self.scheduler_names():
            to_remove = self.delete(scheduler_name, action_name)
            if to_remove:
                removed_scheduler_names.add(to_remove)
        return removed_scheduler_names

    def delete_scheduler(self, scheduler_name: str):
        for sa in self.scheduler_actions:
            if sa.scheduler_name == scheduler_name:
                self.scheduler_actions.remove(sa)
                return scheduler_name
        return None

    def scheduler_names(self):
        return set(sa.scheduler_name for sa in self.scheduler_actions)

    def action_names(self):
        action_names = set()
        for sa in self.scheduler_actions:
            action_names.update(sa.action_names)
        return action_names

    def action_count(self):
        return len(self.action_names())

    def clear(self):
        self.scheduler_actions.clear()


"""
Instances of the namedtuple, DeferredProgram, contain the information necessary
to schedule a program: the time at which to start and the time at which to stop.
"""
DeferredProgram = namedtuple("DeferredProgram", ["program_name", "start", "stop"])


class DeferredPrograms(BaseModel):
    deferred_programs: Set[DeferredProgram] = set()

    def add(self, deferred_program: DeferredProgram):
        self.deferred_programs.add(deferred_program)

    def pop(self):
        filtered = set(dp for dp in self.deferred_programs if dp.start < Now.dt())
        self.deferred_programs = self.deferred_programs - filtered
        return filtered

    def clear_program(self, program_name: str):
        self.deferred_programs = set(
            dp for dp in self.deferred_programs if dp.program_name != program_name
        )

    def clear(self):
        self.deferred_programs = set()

    def count(self):
        return len(self.deferred_programs)


"""
Instances of DatedScheduledActions contains [1] a dictionary having values of
ScheduledActions instances and keys representing datetime instances and [2]
methods that support scheduling within the Dispatcher.

This class has two uses:

1. defer the initiation of schedulers in a ScheduledActions instance to a
   time in the future. [see Dispatcher.deferred_scheduled_actions and
   relevant Dispatcher methods]
2. expire the schedulers in a ScheduledActions instance as of a time in
   the future. [see Dispatcher.expiring_scheduled_actions and relevant
   Dispatcher methods]
"""


class DatedScheduledActions(BaseModel):
    dated_scheduled_actions: Dict[str, ScheduledActions] = {}

    def check_for_dated_actions(
        self,
        schedule_update_thunk: Callable,
        verb: str,
    ):
        """
        This method invokes a supplied thunk when the datetime represented
        in the dictionary key precedes the current time. This thunk expects
        scheduler and action names as arguments.
        """
        now = Now.dt()
        to_remove = []
        for date_time_str in self.dated_scheduled_actions:
            date_time = str_to_dt(date_time_str)
            if date_time < now:
                scheduled_actions = self.dated_scheduled_actions[date_time_str]
                for scheduler_name in scheduled_actions.scheduler_names():
                    for action_name in scheduled_actions.actions(scheduler_name):
                        try:
                            schedule_update_thunk(scheduler_name, action_name)
                        except Exception as exception:
                            logger.error(
                                f"failed to {verb} action ({action_name}) under ({scheduler_name}) as of ({date_time_str})",
                                exception,
                            )
                to_remove.append(date_time_str)
        for date_time_str in to_remove:  # modify outside the previous for-loop
            self.dated_scheduled_actions.pop(date_time_str)

    def apply_date(
        self,
        scheduler_name: str,
        action_name: str,
        date_time: datetime,
    ):
        """
        This method places the scheduler/action pair in the ScheduledActions
        instance corresponding to the supplied datetime.
        """
        date_time_str = dt_to_str(date_time)
        if date_time_str not in self.dated_scheduled_actions:
            self.dated_scheduled_actions[date_time_str] = ScheduledActions()
        scheduled_actions = self.dated_scheduled_actions[date_time_str]
        scheduled_actions.add(scheduler_name, action_name)

    def action_count(self):
        return sum(sa.action_count() for sa in self.dated_scheduled_actions.values())

    def delete_dated_action(self, action_name: str):
        """
        This method deletes all dated ScheduledActions in the dictionary
        that reference the named action.
        """
        date_time_str_to_remove = []
        for date_time_str in self.dated_scheduled_actions:
            scheduled_actions = self.dated_scheduled_actions[date_time_str]
            scheduled_actions.delete_action(action_name)
            if len(scheduled_actions.scheduler_names()) == 0:
                date_time_str_to_remove.append(date_time_str)
        for date_time_str in date_time_str_to_remove:
            self.dated_scheduled_actions.pop(date_time_str)

        """
        This method deletes all dated ScheduledActions in the dictionary
        that reference the named scheduler.
        """

    def delete_dated_scheduler(self, scheduler_name: str):
        date_time_str_to_remove = []
        for date_time_str in self.dated_scheduled_actions:
            scheduled_actions = self.dated_scheduled_actions[date_time_str]
            scheduled_actions.delete_scheduler(scheduler_name)
            if len(scheduled_actions.scheduler_names()) == 0:
                date_time_str_to_remove.append(date_time_str)
        for date_time_str in date_time_str_to_remove:
            self.dated_scheduled_actions.pop(date_time_str)

    def clear(self):
        self.dated_scheduled_actions.clear()



"""
These functions resolve dispatcher element instances from supplied dictionaries.
"""
import logging
from pydantic import ValidationError
from .action import Action, ActionRez
from .scheduler import Scheduler
from .program import Program
from .server import Server
from .util import (
    FilePathe,
    resolve_instance,
    resolve_instance_multi_class,
    Rez,
    DateTime,
    DateTime2,
)


logger = logging.getLogger(__name__)


def resolve_action(dictionary: dict, check_for_found_class: bool = True):
    return resolve_instance(
        Action, dictionary, check_for_found_class=check_for_found_class
    )


def resolve_scheduler(dictionary: dict, check_for_found_class: bool = True):
    return resolve_instance(
        Scheduler, dictionary, check_for_found_class=check_for_found_class
    )


def resolve_program(dictionary: dict, check_for_found_class: bool = True):
    return resolve_instance(
        Program, dictionary, check_for_found_class=check_for_found_class
    )


def resolve_server(dictionary: dict, check_for_found_class: bool = True):
    result = resolve_instance(
        Server, dictionary, check_for_found_class=check_for_found_class
    )
    return result


def resolve_rez(dictionary: dict, check_for_found_class: bool = False):
    result = resolve_instance_multi_class(
        [
            Rez,
            DateTime,
            DateTime2,
            # Action, leave this out for now. Some fastAPI issues.
        ],  # BaseModel objects that may be found in a Rez
        dictionary,
        check_for_found_class=check_for_found_class,
    )
    return result


def resolve_action_rez(dictionary: dict, check_for_found_class: bool = False):
    result = resolve_instance_multi_class(
        [
            Rez,
            Action,
            DateTime,
            DateTime2,
            ActionRez,
        ],  # BaseModel objects that may be found in an ActionRez
        dictionary,
        check_for_found_class=check_for_found_class,
    )
    return result


def resolve_file_pathe(dictionary: dict, check_for_found_class: bool = True):
    try:
        return FilePathe(**dictionary)
    except ValidationError as error:
        if check_for_found_class:
            raise error
        else:
            return dictionary



"""
A server represents a whendo process that listens on a host:port
and responds to restful api requests. Server objects are stored
on a whendo server to facilitate running schedulers and executing
actions on more than one machine.
"""

from pydantic import BaseModel
from typing import Dict, Set, List, Optional
from enum import Enum
import logging
from .util import KeyTagMode


logger = logging.getLogger(__name__)


class Server(BaseModel):
    host: str
    port: int
    tags: Optional[Dict[str, List[str]]] = None

    """
    A server has a dictionary mapping keys to sets of tags (sets
    implemented as lists). Keys serve as a way to qualify tags. It's
    up to the user to use this key:tag-set structure to some benefit.
    """

    def description(self):
        return f"This whendo server listens on host:port ({self.host}:{self.port} with tags ({self.tags})."

    def add_key_tag(self, key: str, tag: str):
        """
        Add a key:tag pair to the tags dictionary.
        """
        self.add_key_tags(key_tags={key: [tag]})

    def add_key_tags(self, key_tags: Dict[str, List[str]]):
        """
        The new tags dictionary will include all supplied
        key:tag pairs.
        """
        for key in key_tags:
            if not self.tags:
                self.tags = {}
            if key in self.tags:
                tags = self.tags[key]
                for tag in key_tags[key]:
                    if tag not in tags:
                        tags.append(tag)
            else:
                self.tags[key] = key_tags[key]

    def has_key(self, key: str):
        return key in self.tags if self.tags else False

    def has_key_tag(self, key: str, tag: str):
        if self.tags:
            return key in self.tags and tag in self.tags[key]
        else:
            return False

    def has_tag(self, tag: str):
        if self.tags:
            return any(tag in self.tags[x] for x in self.tags)
        else:
            return False

    def delete_key_tag(self, key: str, tag: str):
        """
        Remove the tag from the key's set, removing
        an empty key.
        """
        if self.tags:
            tags = self.tags.get(key, None)
            if tags:
                tags.remove(tag)
                if len(tags) == 0:
                    self.tags.pop(key)

    def delete_key(self, key: str):
        """
        Remove the dang key.
        """
        if self.tags:
            self.tags.pop(key)

    def delete_tag(self, tag: str):
        """
        Delete tag from all tag sets and remove
        all consequently empty tag sets.
        """
        if self.tags:
            to_remove: Set[str] = set()
            tags_copy = self.tags.copy()
            for key in tags_copy:
                if tag in self.tags[key]:
                    self.tags[key].remove(tag)
                    if len(self.tags[key]) == 0:
                        to_remove.add(key)
            for key in to_remove:
                self.tags.pop(key)

    def get_tags(self):
        return self.tags

    def get_tags_by_key(self, key: str):
        """
        Return a key's tags.
        """
        return self.tags.get(key, None) if self.tags else []

    def get_keys(self, tags: List[str], key_tag_mode: KeyTagMode = KeyTagMode.ANY):
        if self.tags:
            if key_tag_mode == KeyTagMode.ANY:
                """
                Return the keys having tag sets that intersect the supplied tags.
                """
                return set(
                    key
                    for key in self.tags
                    if len(set(tags).intersection(self.tags[key])) > 0
                )
            if key_tag_mode == KeyTagMode.ALL:
                """
                Return the keys where the supplied tag set covers the key's tags.
                """
                return set(
                    key
                    for key in self.tags
                    if len(set(self.tags[key]).difference(tags)) == 0
                )
        else:
            return []



from fastapi import APIRouter, status, Depends
from whendo.api.shared import return_success, raised_exception, get_dispatcher
from whendo.core.resolver import resolve_action_rez, resolve_action

router = APIRouter(prefix="/execution", tags=["Execution"])


@router.post("", status_code=status.HTTP_200_OK)
def execute_supplied_action(supplied_action=Depends(resolve_action)):
    try:
        assert supplied_action, f"couldn't resolve class for action ({supplied_action})"
        return get_dispatcher(router).execute_supplied_action(
            supplied_action=supplied_action
        )
    except Exception as e:
        raise raised_exception(
            f"failed to directly execute the action ({supplied_action})", e
        )


@router.post("/with_rez", status_code=status.HTTP_200_OK)
def execute_supplied_action_with_rez(action_rez=Depends(resolve_action_rez)):
    """
    The supplied action needs to be passed as an ActionRez
    """
    try:
        assert action_rez, f"couldn't resolve class for action_rez ({action_rez})"
        action = action_rez.action
        rez = action_rez.rez
        return get_dispatcher(router).execute_supplied_action_with_rez(
            supplied_action=action, rez=rez
        )
    except Exception as e:
        raise raised_exception(
            f"failed to directly execute the action embedded in ({action_rez})",
            e,
        )



"""
This script establishes the top FastAPI path and all sub-path routers.
It also initializes the Dispatcher for the calling ASGI server.
"""
from fastapi import FastAPI
from whendo.core.dispatcher import DispatcherSingleton
from whendo.api.router import (
    actions,
    schedulers,
    dispatcher,
    jobs,
    execution,
    programs,
    servers,
)
from whendo.api.shared import set_dispatcher
import whendo.log.init

app = FastAPI()


@app.get("/")
async def root():
    return "whengo API server started"


dispatcher_instance = DispatcherSingleton.get()

app.include_router(set_dispatcher(actions.router, dispatcher_instance))
app.include_router(set_dispatcher(schedulers.router, dispatcher_instance))
app.include_router(set_dispatcher(dispatcher.router, dispatcher_instance))
app.include_router(set_dispatcher(jobs.router, dispatcher_instance))
app.include_router(set_dispatcher(execution.router, dispatcher_instance))
app.include_router(set_dispatcher(programs.router, dispatcher_instance))
app.include_router(set_dispatcher(servers.router, dispatcher_instance))



import logging
import os
import whendo.core.util as util

logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
    },
    "handlers": {
        "default_handler": {
            "class": "logging.FileHandler",
            "level": "INFO",
            "formatter": "standard",
            "filename": os.path.join(util.Dirs.log_dir(), "job.log"),
            "encoding": "utf8",
        },
    },
    "loggers": {
        "": {"handlers": ["default_handler"], "level": "DEBUG", "propagate": False}
    },
}
logging.config.dictConfig(logging_config)



from typing import Optional, Any
import os
import logging
from typing import Optional
from whendo.core.util import PP, Dirs, Now
from whendo.core.action import Action, Rez

logger = logging.getLogger(__name__)


class FileAppend(Action):
    """
    This action appends supplied stuff to a file.
    """

    file_append: str = "file_append"
    file: Optional[str] = None
    relative_to_output_dir: Optional[bool] = None
    payload: Optional[Any] = None
    header: Optional[str] = None
    print_header: Optional[bool] = None

    def description(self):
        return f"This action appends payload ({self.payload}) to file ({self.file})."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        file = flds.get("file", None)
        if file == None:
            raise ValueError("file missing")
        print_header = flds.get("print_header", None)
        if print_header == None:
            print_header = True
        relative_to_output_dir = flds.get("relative_to_output_dir", None)
        if relative_to_output_dir == None:
            relative_to_output_dir = True
        payload = (
            rez.result
            if rez and rez.result is not None
            else flds.get("payload", {"payload": "***PAYLOAD EMPTY***"})
        )
        file = os.path.join(Dirs.output_dir(), file) if relative_to_output_dir else file
        if print_header:
            header_txt = flds.get("header", f"file ({file}) at ({Now.s()})")
            boundary = "".join(["-"] * len(header_txt))
            output_header_txt = f"{boundary}\n{header_txt}\n{boundary}\n"
        with open(file, "a") as outfile:
            if print_header:
                outfile.write(output_header_txt)
            PP.pprint(payload, stream=outfile)
            outfile.write("\n")
        result = f"file ({file}) appended."
        return self.action_result(result=result, rez=rez, flds=rez.flds if rez else {})



import pytest
import time
from datetime import timedelta
from typing import Optional, Dict, Any
from whendo.core.util import Rez, SystemInfo, Now, KeyTagMode, DateTime, Rez
from whendo.core.action import Action
from whendo.core.server import Server
from whendo.core.actions.list_action import (
    UntilFailure,
    All,
    Terminate,
    IfElse,
    RaiseCmp,
    Result,
)
from whendo.core.schedulers.timed_scheduler import Timely
from whendo.core.scheduler import Immediately
from whendo.core.dispatcher import Dispatcher
from whendo.core.programs.simple_program import PBEProgram
from whendo.core.actions.dispatch_action import (
    ScheduleProgram,
    UnscheduleProgram,
    ScheduleAction,
    DeferAction,
    ExpireAction,
)
from whendo.core.timed import Timed
from .fixtures import port, host

pause = 3


def test_server_all_1(friends, servers):
    dispatcher, scheduler, action = friends()
    aqua, teal = servers()
    dispatcher.add_server(server_name="aqua", server=aqua)
    dispatcher.add_server(server_name="teal", server=teal)
    mode = KeyTagMode.ALL
    result = dispatcher.get_servers_by_tags(
        key_tags={"foo": ["bar", "baz"]}, key_tag_mode=mode
    )
    assert len(result) == 2


def test_server_all_2(friends, servers):
    dispatcher, scheduler, action = friends()
    aqua, teal = servers()
    dispatcher.add_server(server_name="aqua", server=aqua)
    dispatcher.add_server(server_name="teal", server=teal)
    mode = KeyTagMode.ALL
    result = dispatcher.get_servers_by_tags(
        key_tags={"foo": ["bar"]}, key_tag_mode=mode
    )
    assert len(result) == 1


def test_server_all_3(friends, servers):
    dispatcher, scheduler, action = friends()
    aqua, teal = servers()
    dispatcher.add_server(server_name="aqua", server=aqua)
    dispatcher.add_server(server_name="teal", server=teal)
    mode = KeyTagMode.ALL
    result = dispatcher.get_servers_by_tags(key_tags={"foo": []}, key_tag_mode=mode)
    assert len(result) == 0


def test_server_all_4(friends, servers):
    dispatcher, scheduler, action = friends()
    aqua, teal = servers()
    dispatcher.add_server(server_name="aqua", server=aqua)
    dispatcher.add_server(server_name="teal", server=teal)
    mode = KeyTagMode.ALL
    result = dispatcher.get_servers_by_tags(
        key_tags={"foo": ["clasp"]}, key_tag_mode=mode
    )
    assert len(result) == 0


def test_server_any_1(friends, servers):
    dispatcher, scheduler, action = friends()
    aqua, teal = servers()
    dispatcher.add_server(server_name="aqua", server=aqua)
    dispatcher.add_server(server_name="teal", server=teal)
    mode = KeyTagMode.ANY
    result = dispatcher.get_servers_by_tags(
        key_tags={"foo": ["bar", "baz"]}, key_tag_mode=mode
    )
    assert len(result) == 2


def test_server_any_2(friends, servers):
    dispatcher, scheduler, action = friends()
    aqua, teal = servers()
    dispatcher.add_server(server_name="aqua", server=aqua)
    dispatcher.add_server(server_name="teal", server=teal)
    mode = KeyTagMode.ANY
    result = dispatcher.get_servers_by_tags(
        key_tags={"foo": ["bar"]}, key_tag_mode=mode
    )
    assert len(result) == 2


def test_server_any_3(friends, servers):
    dispatcher, scheduler, action = friends()
    aqua, teal = servers()
    dispatcher.add_server(server_name="aqua", server=aqua)
    dispatcher.add_server(server_name="teal", server=teal)
    mode = KeyTagMode.ANY
    result = dispatcher.get_servers_by_tags(key_tags={"foo": []}, key_tag_mode=mode)
    assert len(result) == 0


def test_server_any_4(friends, servers):
    dispatcher, scheduler, action = friends()
    aqua, teal = servers()
    dispatcher.add_server(server_name="aqua", server=aqua)
    dispatcher.add_server(server_name="teal", server=teal)
    mode = KeyTagMode.ANY
    result = dispatcher.get_servers_by_tags(
        key_tags={"foo": ["clasp"]}, key_tag_mode=mode
    )
    assert len(result) == 0


def test_schedule_action(friends):
    """
    Tests Dispatcher and Timed objects running a scheduled action.
    """
    dispatcher, scheduler, action = friends()

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")

    assert dispatcher.get_scheduled_action_count() == 1

    dispatcher.run_jobs()
    time.sleep(pause)
    dispatcher.stop_jobs()
    dispatcher.clear_jobs()

    assert action.flea_count > 0


def test_schedule_action_action(friends):
    """
    Tests Dispatcher and Timed objects running a scheduled action.
    """
    dispatcher, scheduler, action = friends()

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    schedule_action = ScheduleAction(scheduler_name="bar", action_name="foo")
    schedule_action.execute()

    assert dispatcher.get_scheduled_action_count() == 1

    dispatcher.run_jobs()
    time.sleep(pause)
    dispatcher.stop_jobs()
    dispatcher.clear_jobs()

    assert action.flea_count > 0


# def test_dispatcher_action_args_1(friends):
#     """
#     Tests computation of args based on fields, data and mode (=field).
#     """
#     dispatcher, scheduler, action = friends()
#     action2 = FleaCount(flea_count=100)

#     dispatcher.add_action("foo", action)
#     dispatcher.add_action("flea", action2)
#     dispatcher.add_scheduler("bar", scheduler)
#     schedule_action = ScheduleAction(
#         scheduler_name="bar", action_name="foo", mode=DispActionMode.FIELD
#     )
#     args = schedule_action.compute_args(
#         args={"scheduler_name": "bar", "action_name": "foo"},
#         data={"action_name": "flea"},
#     )
#     assert args["scheduler_name"] == "bar"
#     assert args["action_name"] == "foo"


# def test_dispatcher_action_args_2(friends):
#     """
#     Tests computation of args based on fields, data and mode (=data).
#     """
#     dispatcher, scheduler, action = friends()
#     action2 = FleaCount(flea_count=100)

#     dispatcher.add_action("foo", action)
#     dispatcher.add_action("flea", action2)
#     dispatcher.add_scheduler("bar", scheduler)
#     schedule_action = ScheduleAction(
#         scheduler_name="bar", action_name="foo", mode=DispActionMode.DATA
#     )
#     args = schedule_action.compute_args(
#         args={"scheduler_name": "bar", "action_name": "foo"},
#         data={"action_name": "flea"},
#     )
#     assert args["scheduler_name"] == "bar"
#     assert args["action_name"] == "flea"


# def test_dispatcher_action_args_3(friends):
#     """
#     Tests computation of args based on fields, data and mode (=field).
#     """
#     dispatcher, scheduler, action = friends()
#     action2 = FleaCount(flea_count=100)

#     dispatcher.add_action("foo", action)
#     dispatcher.add_action("flea", action2)
#     dispatcher.add_scheduler("bar", scheduler)
#     schedule_action = ScheduleAction(
#         scheduler_name="bar", action_name="foo", mode=DispActionMode.FIELD
#     )
#     args = schedule_action.compute_args(
#         args={"scheduler_name": "bar", "action_name": "foo"},
#         data={"result": {"action_name": "flea"}},
#     )
#     assert args["scheduler_name"] == "bar"
#     assert args["action_name"] == "foo"


# def test_dispatcher_action_args_4(friends):
#     """
#     Tests computation of args based on fields, data and mode (=data).
#     """
#     dispatcher, scheduler, action = friends()
#     action2 = FleaCount(flea_count=100)

#     dispatcher.add_action("foo", action)
#     dispatcher.add_action("flea", action2)
#     dispatcher.add_scheduler("bar", scheduler)
#     schedule_action = ScheduleAction(
#         scheduler_name="bar", action_name="foo", mode=DispActionMode.DATA
#     )
#     args = schedule_action.compute_args(
#         args={"scheduler_name": "bar", "action_name": "foo"},
#         data={"result": {"action_name": "flea"}},
#     )
#     assert args["scheduler_name"] == "bar"
#     assert args["action_name"] == "flea"


def test_schedule_action_action_data_1(friends):
    """
    Tests Dispatcher and Timed objects running a scheduled action.
    """
    dispatcher, scheduler, action = friends()
    action2 = FleaCount(flea_count=100)

    dispatcher.add_action("foo", action)
    dispatcher.add_action("flea", action2)
    dispatcher.add_scheduler("bar", scheduler)
    schedule_action = ScheduleAction(scheduler_name="bar", action_name="foo")
    schedule_action.execute(rez=Rez(flds={"action_name": "flea"}))

    assert dispatcher.get_scheduled_action_count() == 1

    dispatcher.run_jobs()
    time.sleep(pause)
    dispatcher.stop_jobs()
    dispatcher.clear_jobs()

    assert action.flea_count > 1
    assert action2.flea_count == 100


def test_schedule_action_action_data_2(friends):
    """
    Tests Dispatcher and Timed objects running a scheduled action.
    """
    dispatcher, scheduler, action = friends()
    action2 = FleaCount(flea_count=100)

    dispatcher.add_action("foo", action)
    dispatcher.add_action("flea", action2)
    dispatcher.add_scheduler("bar", scheduler)
    schedule_action = ScheduleAction(scheduler_name="bar")
    schedule_action.execute(rez=Rez(flds={"action_name": "flea"}))

    assert dispatcher.get_scheduled_action_count() == 1

    dispatcher.run_jobs()
    time.sleep(pause)
    dispatcher.stop_jobs()
    dispatcher.clear_jobs()

    assert action.flea_count == 0
    assert action2.flea_count > 100


def test_unschedule_scheduler(friends):
    """
    Tests unscheduling a scheduler.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")

    assert dispatcher.job_count() == 1

    dispatcher.unschedule_scheduler("bar")

    assert dispatcher.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == dispatcher.job_count()

    # make sure that bar and foo remain
    assert dispatcher.get_scheduler("bar")
    assert dispatcher.get_action("foo")


def test_unschedule_all(friends):
    """
    Tests unscheduling all schedulers.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")
    assert dispatcher.job_count() == 1
    assert dispatcher.get_scheduled_action_count() == 1

    dispatcher.clear_jobs()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo2", action)
    dispatcher.schedule_action("bar", "foo2")
    assert dispatcher.get_scheduled_action_count() == 2

    dispatcher.unschedule_all_schedulers()
    assert dispatcher.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == 0


def test_reschedule_all(friends):
    """
    Tests unscheduling a scheduler.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")
    assert dispatcher.job_count() == 1
    assert dispatcher.get_scheduled_action_count() == 1

    dispatcher.clear_jobs()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo2", action)
    dispatcher.schedule_action("bar", "foo2")
    assert dispatcher.get_scheduled_action_count() == 2

    dispatcher.reschedule_all_schedulers()
    assert dispatcher.job_count() == 1
    assert dispatcher.get_scheduled_action_count() == 2


def test_clear_dispatcher(friends):
    """
    Tests clearing a dispatcher.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")
    assert dispatcher.job_count() == 1

    dispatcher.clear_all()
    assert dispatcher.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == dispatcher.job_count()

    # make sure that bar and foo are Gone
    assert dispatcher.get_scheduler("bar") is None
    assert dispatcher.get_action("foo") is None


def test_scheduled_action_count(friends):
    """
    Tests scheduled action count
    """
    # original
    dispatcher, scheduler, action = friends()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action(action_name="foo", scheduler_name="bar")
    assert 1 == dispatcher.get_scheduled_action_count()
    assert 1 == dispatcher.job_count()


def test_jobs_are_running(friends):
    dispatcher, scheduler, action = friends()
    try:
        dispatcher.run_jobs()
        assert dispatcher.jobs_are_running()
    finally:
        try:
            dispatcher.stop_jobs()
        except:
            pass


def test_jobs_are_not_running(friends):
    dispatcher, scheduler, action = friends()
    try:
        dispatcher.run_jobs()
        assert dispatcher.jobs_are_running()
        dispatcher.stop_jobs()
        assert not dispatcher.jobs_are_running()
    finally:
        try:
            dispatcher.stop_jobs()
        except:
            pass


def test_replace_dispatcher(friends):
    """
    Tests replacing a dispatcher
    """
    # original
    dispatcher, scheduler, action = friends()
    saved_dir = dispatcher.get_saved_dir()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action(action_name="foo", scheduler_name="bar")
    # replacement
    replacement = Dispatcher()  # no saved_dir
    replacement.add_action("flea", action)
    replacement.add_scheduler("bath", scheduler)
    replacement.schedule_action(action_name="flea", scheduler_name="bath")
    # do the business
    dispatcher.replace_all(replacement)
    # is everyone okay?
    assert not dispatcher.get_action("foo")
    assert not dispatcher.get_scheduler("bar")
    assert dispatcher.get_action("flea")
    assert dispatcher.get_scheduler("bath")
    assert {"bath"} == set(k for k in dispatcher.get_schedulers())
    assert {"flea"} == set(k for k in dispatcher.get_actions())
    assert {"bath"} == set(
        k for k in dispatcher.get_scheduled_actions().scheduler_names()
    )
    assert {"flea"} == dispatcher.get_scheduled_actions().actions("bath")


def test_load_dispatcher(friends):
    """
    Tests loading a dispatcher
    """
    dispatcher, scheduler, action = friends()
    saved_dir = dispatcher.get_saved_dir()
    assert saved_dir is not None
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action(action_name="foo", scheduler_name="bar")
    dispatcher2 = dispatcher.load_current()
    assert dispatcher2 is not None, f"saved_dir({saved_dir})"
    assert set(k for k in dispatcher.get_actions()) == set(
        k for k in dispatcher2.get_actions()
    )
    assert set(k for k in dispatcher.get_schedulers()) == set(
        k for k in dispatcher2.get_schedulers()
    )
    assert set(k for k in dispatcher.get_scheduled_actions().action_names()) == set(
        k for k in dispatcher2.get_scheduled_actions().action_names()
    )


def test_saved_dir_1(tmp_path):
    saved_dir = str(tmp_path)
    dispatcher = Dispatcher()
    dispatcher.set_saved_dir(saved_dir=saved_dir)
    assert dispatcher.get_saved_dir() == saved_dir


def test_saved_dir_2(tmp_path):
    saved_dir = str(tmp_path)
    dispatcher = Dispatcher(saved_dir=saved_dir)
    assert dispatcher.get_saved_dir() == saved_dir


def test_defer_action(friends):
    """
    Want to observe the scheduling move from deferred state to ready state.
    """
    dispatcher, scheduler, action = friends()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)

    assert 0 == dispatcher.get_deferred_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    dispatcher.defer_action(
        scheduler_name="bar", action_name="foo", wait_until=Now.dt()
    )

    # deferred state -- can run jobs and actions will _not_ be executed
    assert 1 == dispatcher.get_deferred_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    time.sleep(6)  # the out-of-band job runs every five seconds

    # ready state -- can run jobs and actions will be executed
    assert 0 == dispatcher.get_deferred_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()


def test_defer_action_action(friends):
    """
    Want to observe the scheduling move from deferred state to ready state.
    """
    dispatcher, scheduler, action = friends()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)

    assert 0 == dispatcher.get_deferred_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    defer_action = DeferAction(
        scheduler_name="bar",
        action_name="foo",
        wait_until=DateTime(dt=Now.dt()),
    )
    defer_action.execute()

    # deferred state -- can run jobs and actions will _not_ be executed
    assert 1 == dispatcher.get_deferred_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    time.sleep(6)  # the out-of-band job runs every five seconds

    # ready state -- can run jobs and actions will be executed
    assert 0 == dispatcher.get_deferred_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()


def test_expire_action(friends):
    """
    Want to observe the scheduling move from deferred state to ready state.
    """
    dispatcher, scheduler, action = friends()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)

    assert 0 == dispatcher.get_expiring_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    dispatcher.schedule_action(scheduler_name="bar", action_name="foo")

    assert 0 == dispatcher.get_expiring_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()

    dispatcher.expire_action(
        scheduler_name="bar",
        action_name="foo",
        expire_on=Now.dt() + timedelta(seconds=1),
    )

    assert 1 == dispatcher.get_expiring_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()

    time.sleep(6)  # the out-of-band job runs every 2-5 seconds

    assert 0 == dispatcher.get_expiring_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()


def test_expire_action_action(friends):
    """
    Want to observe the scheduling move from deferred state to ready state.
    """
    dispatcher, scheduler, action = friends()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)

    assert 0 == dispatcher.get_expiring_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    dispatcher.schedule_action(scheduler_name="bar", action_name="foo")

    assert 0 == dispatcher.get_expiring_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()

    expire_action = ExpireAction(
        scheduler_name="bar",
        action_name="foo",
        expire_on=DateTime(dt=Now.dt() + timedelta(seconds=2)),
    )
    expire_action.execute()

    assert 1 == dispatcher.get_expiring_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()

    time.sleep(6)  # the out-of-band job runs every 2-5 seconds

    assert 0 == dispatcher.get_expiring_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()


def test_immediately(friends):
    """
    Want to observe that action get executed immediately and that schedulers_actions
    is not impacted.
    """

    dispatcher, scheduler, action = friends()

    class TestAction(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result(result=self.fleas)

    test_action = TestAction()

    assert dispatcher.get_scheduled_action_count() == 0
    assert test_action.fleas == 0

    dispatcher.add_action("foo", test_action)
    dispatcher.add_scheduler("imm", Immediately())
    dispatcher.schedule_action(scheduler_name="imm", action_name="foo")

    assert dispatcher.get_scheduled_action_count() == 0
    assert test_action.fleas == 1


def test_program_1(friends):
    """
    Want to observe that a Program's actions are executed.
    """

    dispatcher, scheduler, action = friends()

    class TestAction1(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    class TestAction2(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    class TestAction3(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    action1 = TestAction1()
    action2 = TestAction2()
    action3 = TestAction3()

    dispatcher.add_action("foo1", action1)
    dispatcher.add_action("foo2", action2)
    dispatcher.add_action("foo3", action3)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.add_scheduler("immediately", Immediately())

    program = PBEProgram().prologue("foo1").body_element("bar", "foo2").epilogue("foo3")
    dispatcher.add_program("baz", program)

    start = Now().dt() + timedelta(seconds=1)
    stop = start + timedelta(seconds=4)

    dispatcher.run_jobs()
    dispatcher.schedule_program("baz", start, stop)

    assert action1.fleas == 0
    time.sleep(3)
    assert action1.fleas == 1
    time.sleep(4)
    assert action2.fleas >= 2
    time.sleep(2)
    assert action3.fleas == 1


def test_unschedule_program(friends):
    """
    Want to observe that a Program's actions are not executed
    after being unscheduled prior to the deferral time.
    """

    dispatcher, scheduler, action = friends()

    class TestAction1(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    class TestAction2(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    class TestAction3(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    action1 = TestAction1()
    action2 = TestAction2()
    action3 = TestAction3()

    dispatcher.add_action("foo1", action1)
    dispatcher.add_action("foo2", action2)
    dispatcher.add_action("foo3", action3)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.add_scheduler("immediately", Immediately())

    program = PBEProgram().prologue("foo1").body_element("bar", "foo2").epilogue("foo3")
    dispatcher.add_program("baz", program)
    start = Now().dt() + timedelta(seconds=4)
    stop = start + timedelta(seconds=4)

    dispatcher.run_jobs()
    dispatcher.schedule_program("baz", start, stop)
    assert dispatcher.get_deferred_program_count() == 1
    assert dispatcher.get_scheduled_action_count() == 0

    dispatcher.unschedule_program("baz")
    assert len(dispatcher.get_programs()) == 1
    print("BLEEEEEE", dispatcher.get_deferred_programs())
    assert dispatcher.get_deferred_program_count() == 0

    assert action1.fleas == 0
    time.sleep(3)
    assert action1.fleas == 0
    time.sleep(4)
    assert action2.fleas == 0
    time.sleep(2)
    assert action3.fleas == 0


def test_unschedule_program_action(friends):
    """
    Want to observe that a Program's actions are not executed
    after being unscheduled prior to the deferral time.
    """

    dispatcher, scheduler, action = friends()

    class TestAction1(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    class TestAction2(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    class TestAction3(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    action1 = TestAction1()
    action2 = TestAction2()
    action3 = TestAction3()

    dispatcher.add_action("foo1", action1)
    dispatcher.add_action("foo2", action2)
    dispatcher.add_action("foo3", action3)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.add_scheduler("immediately", Immediately())

    program = PBEProgram().prologue("foo1").body_element("bar", "foo2").epilogue("foo3")
    dispatcher.add_program("baz", program)
    start = Now().dt() + timedelta(seconds=4)
    stop = start + timedelta(seconds=4)

    dispatcher.run_jobs()
    dispatcher.schedule_program("baz", start, stop)
    assert dispatcher.get_deferred_program_count() == 1
    assert dispatcher.get_scheduled_action_count() == 0

    unschedule_program = UnscheduleProgram(program_name="baz")
    unschedule_program.execute()
    time.sleep(1)

    assert len(dispatcher.get_programs()) == 1
    assert dispatcher.get_deferred_program_count() == 0

    assert action1.fleas == 0
    time.sleep(3)
    assert action1.fleas == 0
    time.sleep(4)
    assert action2.fleas == 0
    time.sleep(2)
    assert action3.fleas == 0


def test_delete_program(friends):
    """
    Want to observe that a Program's actions are not executed
    after being deleted prior to the deferral time.
    """

    dispatcher, scheduler, action = friends()

    class TestAction1(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    class TestAction2(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    class TestAction3(Action):
        fleas: int = 0

        def execute(self, tag: str = None, rez: Rez = None):
            self.fleas += 1
            return self.action_result()

    action1 = TestAction1()
    action2 = TestAction2()
    action3 = TestAction3()

    dispatcher.add_action("foo1", action1)
    dispatcher.add_action("foo2", action2)
    dispatcher.add_action("foo3", action3)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.add_scheduler("immediately", Immediately())

    program = PBEProgram().prologue("foo1").body_element("bar", "foo2").epilogue("foo3")
    dispatcher.add_program("baz", program)
    start = Now().dt() + timedelta(seconds=4)
    stop = start + timedelta(seconds=4)
    assert len(dispatcher.get_programs()) == 1

    dispatcher.run_jobs()
    dispatcher.schedule_program("baz", start, stop)
    assert dispatcher.get_deferred_program_count() == 1
    assert dispatcher.get_scheduled_action_count() == 0

    dispatcher.delete_program("baz")
    assert len(dispatcher.get_programs()) == 0
    assert dispatcher.get_deferred_program_count() == 0

    assert action1.fleas == 0
    time.sleep(3)
    assert action1.fleas == 0
    time.sleep(4)
    assert action2.fleas == 0
    time.sleep(2)
    assert action3.fleas == 0


def test_execute_with_rez(friends):
    """
    Want to see execute work with supplied dictionary.
    """
    dispatcher, scheduler, action = friends()
    result = action.execute(rez=Rez(result={"fleacount": "infinite"}))
    assert result.rez.result == {"fleacount": "infinite"}


def test_terminate_scheduler(friends):
    """
    Want to terminate scheduler using TerminateScheduler action.
    """
    dispatcher, scheduler, action = friends()
    action2 = Terminate()
    dispatcher.add_action("foo", action)
    dispatcher.add_action("terminate", action2)
    dispatcher.add_scheduler("bar", scheduler)

    dispatcher.run_jobs()
    dispatcher.schedule_action("bar", "foo")
    time.sleep(2)
    assert action.flea_count >= 1
    assert dispatcher.get_scheduled_action_count() == 1
    dispatcher.schedule_action("bar", "terminate")
    assert dispatcher.get_scheduled_action_count() == 2
    time.sleep(2)
    assert dispatcher.get_scheduled_action_count() == 0


def test_terminate_scheduler_and(friends):
    """
    Want to terminate scheduler using TerminateScheduler action.
    """
    dispatcher, scheduler, action = friends()
    action2 = FleaCount(flea_count=100)
    actions = [action, Terminate(), action2]
    action3 = (
        UntilFailure()
    )  # add actions on next line to use them directly below; pydantic deep copies field values
    action3.actions = actions
    dispatcher.add_action("foo", action3)
    dispatcher.add_scheduler("bar", scheduler)

    dispatcher.run_jobs()
    dispatcher.schedule_action("bar", "foo")
    time.sleep(3)
    assert action.flea_count == 1
    assert action2.flea_count == 100


def test_if_else_1(friends):
    """
    Want to terminate scheduler using TerminateScheduler action.
    """
    dispatcher, scheduler, action = friends()
    action2 = FleaCount(flea_count=100)
    immediately = Immediately()
    dispatcher.add_action("foo1", action)
    dispatcher.add_action("foo2", action2)
    dispatcher.add_scheduler("immediately", immediately)

    if_else = IfElse(
        test_action=RaiseCmp(value=1),
        if_action=ScheduleAction(scheduler_name="immediately", action_name="foo1"),
        else_action=ScheduleAction(scheduler_name="immediately", action_name="foo2"),
    )
    schedule_action = All(actions=[Result(value=2), if_else])

    dispatcher.add_action("schedule_action", schedule_action)
    dispatcher.schedule_action("immediately", "schedule_action")

    assert action.flea_count == 1
    assert action2.flea_count == 100


def test_if_else_2(friends):
    """
    Want to terminate scheduler using TerminateScheduler action.
    """
    dispatcher, scheduler, action = friends()
    action2 = FleaCount(flea_count=100)
    immediately = Immediately()
    dispatcher.add_action("foo1", action)
    dispatcher.add_action("foo2", action2)
    dispatcher.add_scheduler("immediately", immediately)

    if_else = IfElse(
        test_action=RaiseCmp(value=2),
        if_action=ScheduleAction(scheduler_name="immediately", action_name="foo1"),
        else_action=ScheduleAction(scheduler_name="immediately", action_name="foo2"),
    )
    schedule_action = All(actions=[Result(value=2), if_else])

    dispatcher.add_action("schedule_action", schedule_action)
    dispatcher.schedule_action("immediately", "schedule_action")

    assert action.flea_count == 0
    assert action2.flea_count == 101


# ====================================


class FleaCount(Action):
    flea_count: int = 0
    data: Optional[Dict[Any, Any]]

    def execute(self, tag: str = None, rez: Rez = None):
        self.flea_count += 1
        return self.action_result(
            result=self.flea_count, rez=rez, flds=rez.flds if rez else {}
        )


@pytest.fixture
def friends(tmp_path, host, port):
    """ returns a tuple of useful test objects """
    SystemInfo.init(host, port)

    def stuff():
        # want a fresh tuple from the fixture
        dispatcher = Dispatcher(saved_dir=str(tmp_path))
        dispatcher.set_timed(Timed())
        dispatcher.initialize()
        action = FleaCount()
        scheduler = Timely(interval=1)

        return dispatcher, scheduler, action

    return stuff


@pytest.fixture
def servers():
    def stuff():
        server1 = Server(host="localhost", port=8000)
        server1.add_key_tag("foo", "bar")
        server1.add_key_tag("foo", "baz")
        server1.add_key_tag("fleas", "standdown")
        server1.add_key_tag("krimp", "kramp")
        server2 = Server(host="localhost", port=8000)
        server2.add_key_tag("foo", "bar")
        server2.add_key_tag("fleas", "riseup")
        server2.add_key_tag("slip", "slide")
        return (server1, server2)

    return stuff



from pydantic import BaseModel, PrivateAttr
import requests
import logging
from typing import Optional
from whendo.core.action import Action, ActionRez, Rez, RezDict
from whendo.core.scheduler import Scheduler
from whendo.core.server import Server
from whendo.core.resolver import (
    resolve_action,
    resolve_scheduler,
    resolve_file_pathe,
    resolve_program,
    resolve_server,
    resolve_rez,
)
from whendo.core.util import FilePathe, DateTime, Http, DateTime2, Rez
from whendo.core.dispatcher import Dispatcher
from whendo.core.program import Program

logger = logging.getLogger(__name__)


class Client(BaseModel):
    host: str
    port: int
    client_used: bool = False
    http_instance: Optional[Http] = None

    def get_host(self):
        return self.host

    def http(self):
        """
        Cache the Http object.
        """
        if self.http_instance == None:
            self.http_instance = Http(host=self.host, port=self.port)
        return self.http_instance

    # /dispatcher
    def load_dispatcher(self):
        return Dispatcher.resolve(self.http().get(f"/dispatcher/load"))

    def save_dispatcher(self):
        return self.http().get("/dispatcher/save")

    def clear_dispatcher(self):
        return self.http().get("/dispatcher/clear")

    def load_dispatcher_from_name(self, name: str):
        return Dispatcher.resolve(self.http().get(f"/dispatcher/load_from_name/{name}"))

    def save_data_to_name(self, name: str):
        return self.http().get(f"/dispatcher/save_to_name/{name}")

    def get_saved_dir(self):
        return resolve_file_pathe(self.http().get("/dispatcher/saved_dir"))

    def set_saved_dir(self, saved_dir: FilePathe):
        return self.http().put("/dispatcher/saved_dir", saved_dir)

    def replace_dispatcher(self, replacement: Dispatcher):
        return self.http().put("/dispatcher/replace", replacement)

    def describe_all(self):
        return self.http().get("/dispatcher/describe_all")

    # /execution

    def execute_supplied_action(self, supplied_action: Action):
        response = self.http().post(f"/execution", supplied_action)
        return resolve_rez(response)

    def execute_supplied_action_with_rez(self, supplied_action: Action, rez: Rez):
        action_rez = ActionRez(action=supplied_action, rez=rez)
        response = self.http().post(f"/execution/with_rez", action_rez)
        return resolve_rez(response)

    # /actions
    def get_action(self, action_name: str):
        return resolve_action(self.http().get(f"/actions/{action_name}"))

    def get_actions(self):
        actions = self.http().get(f"/actions")
        return {name: resolve_action(actions[name]) for name in actions}

    def describe_action(self, action_name: str):
        return self.http().get(f"/actions/{action_name}/describe")

    def add_action(self, action_name: str, action: Action):
        return self.http().post(f"/actions/{action_name}", action)

    def set_action(self, action_name: str, action: Action):
        return self.http().put(f"/actions/{action_name}", action)

    def delete_action(self, action_name: str):
        return self.http().delete(f"/actions/{action_name}")

    def execute_action(self, action_name: str):
        return resolve_rez(self.http().get(f"/actions/{action_name}/execute"))

    def execute_action_with_rez(self, action_name: str, rez: Rez):
        return resolve_rez(self.http().post(f"/actions/{action_name}/execute", rez))

    # /schedulers

    def schedule_action(self, scheduler_name: str, action_name: str):
        return self.http().get(f"/schedulers/{scheduler_name}/actions/{action_name}")

    def get_schedulers(self):
        schedulers = self.http().get(f"/schedulers")
        return {name: resolve_scheduler(schedulers[name]) for name in schedulers}

    def get_scheduler(self, scheduler_name: str):
        return resolve_scheduler(self.http().get(f"/schedulers/{scheduler_name}"))

    def describe_scheduler(self, scheduler_name: str):
        return self.http().get(f"/schedulers/{scheduler_name}/describe")

    def add_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        return self.http().post(f"/schedulers/{scheduler_name}", scheduler)

    def set_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        return self.http().put(f"/schedulers/{scheduler_name}", scheduler)

    def delete_scheduler(self, scheduler_name: str):
        return self.http().delete(f"/schedulers/{scheduler_name}")

    def unschedule_scheduler(self, scheduler_name: str):
        return self.http().get(f"/schedulers/{scheduler_name}/unschedule")

    def unschedule_all_schedulers(self):
        return self.http().get(f"/schedulers/unschedule_all")

    def unschedule_scheduler_action(self, scheduler_name: str, action_name: str):
        return self.http().get(
            f"/schedulers/{scheduler_name}/actions/{action_name}/unschedule"
        )

    def rescheduler_scheduler(self, scheduler_name: str):
        return self.http().get(f"/schedulers/{scheduler_name}/reschedule")

    def reschedule_all_schedulers(self):
        return self.http().get(f"/schedulers/reschedule_all")

    def scheduled_action_count(self):
        return self.http().get("/schedulers/action_count")

    def clear_all_scheduling(self):
        return self.http().get("/schedulers/clear_scheduling")

    # programs
    def get_program(self, program_name: str):
        return resolve_program(self.http().get(f"/programs/{program_name}"))

    def get_programs(self):
        programs = self.http().get(f"/programs")
        return {name: resolve_program(programs[name]) for name in programs}

    def describe_program(self, program_name: str):
        return self.http().get(f"/programs/{program_name}/describe")

    def add_program(self, program_name: str, program: Program):
        return self.http().post(f"/programs/{program_name}", program)

    def set_program(self, program_name: str, program: Program):
        return self.http().put(f"/programs/{program_name}", program)

    def delete_program(self, program_name: str):
        return self.http().delete(f"/programs/{program_name}")

    def schedule_program(self, program_name: str, start_stop: DateTime2):
        return self.http().post(f"/programs/{program_name}/schedule", start_stop)

    def unschedule_program(self, program_name: str):
        return self.http().get(f"/programs/{program_name}/unschedule")

    def clear_deferred_programs(self):
        return self.http().get(f"/programs/clear_deferred_programs")

    def deferred_program_count(self):
        return self.http().get(f"/programs/deferred_program_count")

    # servers

    def get_server(self, server_name: str):
        return resolve_server(self.http().get(f"/servers/{server_name}"))

    def get_servers(self):
        servers = self.http().get(f"/servers")
        return {name: resolve_server(servers[name]) for name in servers}

    def describe_server(self, server_name: str):
        return self.http().get(f"/servers/{server_name}/describe")

    def add_server(self, server_name: str, server: Server):
        return self.http().post(f"/servers/{server_name}", server)

    def set_server(self, server_name: str, server: Server):
        return self.http().put(f"/servers/{server_name}", server)

    def delete_server(self, server_name: str):
        return self.http().delete(f"/servers/{server_name}")

    def get_server_tags(self, server_name: str):
        return self.http().get(f"/servers/{server_name}/get_tags")

    def get_servers_by_tags(self, mode: str, key_tags: dict):
        servers = self.http().post_dict(f"/servers/by_tags/{mode}", key_tags)
        return [resolve_server(server) for server in servers]

    def execute_on_server(self, server_name: str, action_name: str):
        return resolve_rez(
            self.http().get(f"/servers/{server_name}/actions/{action_name}/execute")
        )

    def execute_on_server_with_rez(self, server_name: str, action_name: str, rez: Rez):
        return resolve_rez(
            self.http().post(
                f"/servers/{server_name}/actions/{action_name}/execute", rez
            )
        )

    def execute_on_servers(self, mode: str, action_name: str, key_tags: dict):
        return self.http().post_dict(
            f"/servers/by_tags/{mode}/actions/{action_name}/execute", key_tags
        )

    def execute_on_servers_with_rez(
        self, mode: str, action_name: str, key_tags: dict, rez: Rez
    ):
        rez_dict = RezDict(r=rez, d=key_tags)
        return self.http().post(
            f"/servers/by_tags/{mode}/actions/{action_name}/execute_with_rez",
            rez_dict,
        )

    # deferrals and expirations
    def defer_action(self, scheduler_name: str, action_name: str, wait_until: DateTime):
        return self.http().post(
            f"/schedulers/{scheduler_name}/actions/{action_name}/defer", wait_until
        )

    def clear_deferred_actions(self):
        return self.http().get(f"/schedulers/clear_deferred_actions")

    def deferred_action_count(self):
        return self.http().get("/schedulers/deferred_action_count")

    def expire_action(self, scheduler_name: str, action_name: str, expire_on: DateTime):
        return self.http().post(
            f"/schedulers/{scheduler_name}/actions/{action_name}/expire", expire_on
        )

    def clear_expiring_actions(self):
        return self.http().get(f"/schedulers/clear_expiring_actions")

    def expiring_action_count(self):
        return self.http().get("/schedulers/expiring_action_count")

    # /jobs
    def run_jobs(self):
        return self.http().get(f"/jobs/run")

    def stop_jobs(self):
        return self.http().get(f"/jobs/stop")

    def jobs_are_running(self):
        return self.http().get(f"/jobs/are_running")

    def job_count(self):
        return self.http().get(f"/jobs/count")

    def clear_jobs(self):
        return self.http().get(f"/jobs/clear")



from pydantic import BaseModel
from httpx import AsyncClient
import json
from whendo.core.action import Action, ActionRez, RezDict
from whendo.core.scheduler import Scheduler
from whendo.core.server import Server
from whendo.core.resolver import (
    resolve_action,
    resolve_scheduler,
    resolve_file_pathe,
    resolve_program,
    resolve_server,
    resolve_rez,
)
from whendo.core.util import FilePathe, DateTime, DateTime2, Rez
from whendo.core.dispatcher import Dispatcher
from whendo.core.program import Program


class ClientAsync(BaseModel):
    """
    This class is very similar to Client, except that the methods
    that typically return BaseModel objects will return supplied
    dictionaries if resolution fails.
    """

    host: str
    port: int

    # /dispatcher
    async def load_dispatcher(self):
        return Dispatcher.resolve(await self.get_as_json(f"/dispatcher/load"))

    async def save_dispatcher(self):
        return await self.get("/dispatcher/save")

    async def clear_dispatcher(self):
        return await self.get("/dispatcher/clear")

    async def load_dispatcher_from_name(self, name: str):
        return Dispatcher.resolve(
            await self.get_as_json(f"/dispatcher/load_from_name/{name}")
        )

    async def save_data_to_name(self, name: str):
        return await self.get(f"/dispatcher/save_to_name/{name}")

    async def get_saved_dir(self):
        return resolve_file_pathe(
            await self.get_as_json("/dispatcher/saved_dir"), check_for_found_class=False
        )

    async def set_saved_dir(self, saved_dir: FilePathe):
        return await self.put("/dispatcher/saved_dir", saved_dir)

    async def replace_dispatcher(self, replacement: Dispatcher):
        return await self.put("/dispatcher/replace", replacement)

    async def describe_all(self):
        return await self.get("/dispatcher/describe_all")

    # /execution
    async def execute_supplied_action(self, supplied_action: Action):
        response = await self.post("/execution", supplied_action)
        return resolve_rez(response.json())

    async def execute_supplied_action_with_rez(self, supplied_action: Action, rez: Rez):
        action_rez = ActionRez(action=supplied_action, rez=rez)
        response = await self.post("/execution/with_rez", action_rez)
        return resolve_rez(response.json())

    # /actions
    async def get_action(self, action_name: str):
        return resolve_action(
            await self.get_as_json(f"/actions/{action_name}"),
            check_for_found_class=False,
        )

    async def get_actions(self):
        actions = await self.get_as_json(f"/actions")
        return {
            name: resolve_action(actions[name], check_for_found_class=False)
            for name in actions
        }

    async def add_action(self, action_name: str, action: Action):
        return await self.post(f"/actions/{action_name}", action)

    async def set_action(self, action_name: str, action: Action):
        return await self.put(f"/actions/{action_name}", action)

    async def delete_action(self, action_name: str):
        return await self.delete(f"/actions/{action_name}")

    async def execute_action(self, action_name: str):
        return await self.get(f"/actions/{action_name}/execute")

    async def execute_action_with_rez(self, action_name: str, rez: Rez):
        return await self.post(f"/actions/{action_name}/execute", rez)

    # /schedulers
    async def schedule_action(self, scheduler_name: str, action_name: str):
        return await self.get(f"/schedulers/{scheduler_name}/actions/{action_name}")

    async def get_scheduler(self, scheduler_name: str):
        return resolve_scheduler(
            await self.get_as_json(f"/schedulers/{scheduler_name}"),
            check_for_found_class=False,
        )

    async def get_schedulers(self):
        schedulers = await self.get_as_json(f"/schedulers")
        return {
            name: resolve_scheduler(schedulers[name], check_for_found_class=False)
            for name in schedulers
        }

    async def describe_scheduler(self, scheduler_name: str):
        return await self.get(f"/schedulers/{scheduler_name}/describe")

    async def add_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        return await self.post(f"/schedulers/{scheduler_name}", scheduler)

    async def set_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        return await self.put(f"/schedulers/{scheduler_name}", scheduler)

    async def delete_scheduler(self, scheduler_name: str):
        return await self.delete(f"/schedulers/{scheduler_name}")

    async def unschedule_scheduler(self, scheduler_name: str):
        return await self.get(f"/schedulers/{scheduler_name}/unschedule")

    async def unschedule_all_schedulers(self):
        return await self.get(f"/schedulers/unschedule_all")

    async def reschedule_all_schedulers(self):
        return await self.get(f"/schedulers/reschedule_all")

    async def execute_scheduled_actions(self, scheduler_name: str):
        return await self.get(f"/schedulers/{scheduler_name}/execute")

    async def scheduled_action_count(self):
        return await self.get("/schedulers/action_count")

    async def clear_all_scheduling(self):
        return await self.get("/schedulers/clear_scheduling")

    # programs
    async def get_program(self, program_name: str):
        return resolve_program(
            await self.get_as_json(f"/programs/{program_name}"),
            check_for_found_class=False,
        )

    async def get_programs(self):
        programs = await self.get_as_json(f"/programs")
        return {
            name: resolve_program(programs[name], check_for_found_class=False)
            for name in programs
        }

    async def add_program(self, program_name: str, program: Program):
        return await self.post(f"/programs/{program_name}", program)

    async def set_program(self, program_name: str, program: Program):
        return await self.put(f"/programs/{program_name}", program)

    async def delete_program(self, program_name: str):
        return await self.delete(f"/programs/{program_name}")

    async def schedule_program(self, program_name: str, start_stop: DateTime2):
        return await self.post(f"/programs/{program_name}/schedule", start_stop)

    async def unschedule_program(self, program_name: str):
        return await self.get(f"/programs/{program_name}/unschedule")

    async def clear_deferred_programs(self):
        return await self.get(f"/programs/clear_deferred_programs")

    async def deferred_program_count(self):
        return await self.get(f"/programs/deferred_program_count")

    # servers

    async def get_server(self, server_name: str):
        return resolve_server(
            await self.get_as_json(f"/servers/{server_name}"),
            check_for_found_class=False,
        )

    async def get_servers(self):
        servers = await self.get_as_json(f"/servers")
        return {
            name: resolve_server(servers[name], check_for_found_class=False)
            for name in servers
        }

    async def describe_server(self, server_name: str):
        return await self.get(f"/servers/{server_name}/describe")

    async def add_server(self, server_name: str, server: Server):
        return await self.post(f"/servers/{server_name}", server)

    async def set_server(self, server_name: str, server: Server):
        return await self.put(f"/servers/{server_name}", server)

    async def delete_server(self, server_name: str):
        return await self.delete(f"/servers/{server_name}")

    async def get_server_tags(self, server_name: str):
        return await self.get(f"/servers/{server_name}/get_tags")

    async def get_servers_by_tags(self, server_name: str, key_tags: dict):
        return await self.post_dict(f"/servers/by_tags", key_tags)

    async def execute_on_server(self, server_name: str, action_name: str):
        return resolve_rez(
            await self.get(f"/servers/{server_name}/actions/{action_name}/execute")
        )

    async def execute_on_server_with_rez(
        self, server_name: str, action_name: str, rez: Rez
    ):
        return resolve_rez(
            await self.post(
                f"/servers/{server_name}/actions/{action_name}/execute", rez
            )
        )

    async def execute_on_servers(self, mode: str, action_name: str, key_tags: dict):
        return await self.post_dict(
            f"/servers/by_tags/{mode}/actions/{action_name}/execute", key_tags
        )

    async def execute_on_servers_with_rez(
        self, mode: str, action_name: str, key_tags: dict, rez: Rez
    ):
        rez_dict = RezDict(r=rez, d=key_tags)
        return await self.post(
            f"/servers/by_tags/{mode}/actions/{action_name}/execute_with_rez",
            rez_dict,
        )

    # deferrals and expirations
    async def defer_action(
        self, scheduler_name: str, action_name: str, wait_until: DateTime
    ):
        return await self.post(
            f"/schedulers/{scheduler_name}/actions/{action_name}/defer", wait_until
        )

    async def clear_deferred_actions(self):
        return await self.get(f"/schedulers/clear_deferred_actions")

    async def deferred_action_count(self):
        return await self.get("/schedulers/deferred_action_count")

    async def expire_action(
        self, scheduler_name: str, action_name: str, expire_on: DateTime
    ):
        return await self.post(
            f"/schedulers/{scheduler_name}/actions/{action_name}/expire", expire_on
        )

    async def clear_expiring_actions(self):
        return await self.get(f"/schedulers/clear_expiring_actions")

    async def expiring_action_count(self):
        return await self.get("/schedulers/expiring_action_count")

    # /jobs
    async def run_jobs(self):
        return await self.get(f"/jobs/run")

    async def stop_jobs(self):
        return await self.get(f"/jobs/stop")

    async def jobs_are_running(self):
        return await self.get(f"/jobs/are_running")

    async def job_count(self):
        return await self.get(f"/jobs/count")

    async def clear_jobs(self):
        return await self.get(f"/jobs/clear")

    # verbs
    async def get(self, path: str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.get(url=path)

    async def get_as_json(self, path: str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            response = await ac.get(url=path)
            return response.json()

    async def put(self, path: str, data: BaseModel):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.put(url=path, data=data.json())

    async def post(self, path: str, data: BaseModel):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.post(url=path, data=data.json())

    async def post_dict(self, path: str, data: dict):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.post(url=path, data=json.dumps(data))

    async def delete(self, path: str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.delete(url=path)

    def base_url(self):
        return f"http://{self.host}:{self.port}"



"""
This script establishes the top FastAPI path and all sub-path routers.
It also initializes the Dispatcher for the calling ASGI server.

This version is only used in unit testing -- the Dispatcher construction
is different from main.py.
"""
from fastapi import FastAPI
from whendo.core.dispatcher import Dispatcher
from whendo.core.timed import Timed
from whendo.api.router import (
    actions,
    schedulers,
    dispatcher,
    jobs,
    execution,
    programs,
    servers,
)
from whendo.api.shared import set_dispatcher

app = FastAPI()


@app.get("/")
async def root():
    return "whengo API server started (unit test)"


dispatcher_instance = Dispatcher()
dispatcher_instance.set_timed(Timed())
dispatcher_instance.initialize()

app.include_router(set_dispatcher(actions.router, dispatcher_instance))
app.include_router(set_dispatcher(schedulers.router, dispatcher_instance))
app.include_router(set_dispatcher(dispatcher.router, dispatcher_instance))
app.include_router(set_dispatcher(jobs.router, dispatcher_instance))
app.include_router(set_dispatcher(execution.router, dispatcher_instance))
app.include_router(set_dispatcher(programs.router, dispatcher_instance))
app.include_router(set_dispatcher(servers.router, dispatcher_instance))



from pydantic import BaseModel
from typing import Dict, Any, Optional
import logging
from .util import object_info, SystemInfo, Rez


logger = logging.getLogger(__name__)


class Action(BaseModel):
    """
    Actions get something done.
    """

    def description(self):
        return "This has no description."

    def execute(self, tag: str = None, rez: Rez = None) -> Rez:
        return self.action_result(rez=rez)

    def info(self):
        return object_info(self)

    def flat(self):
        return self.json()

    def local_host(self):
        return SystemInfo.get()["host"]

    def local_port(self):
        return SystemInfo.get()["port"]

    def local_time(self):
        return SystemInfo.get()["current"]

    def local_info(self):
        return {
            "host": self.local_host(),
            "port": self.local_port(),
            "time": self.local_time(),
        }

    def fields(self):
        """
        Return only non-meta fields.
        """
        return set(
            name for name in self.__fields__ if self.__fields__[name].default != name
        )

    def field_values(self):
        return {
            f: self.__dict__[f] for f in self.fields() if self.__dict__[f] is not None
        }

    def compute_flds(self, rez: Rez = None):
        field_values = self.field_values()
        if rez and rez.flds:
            rez_flds = rez.flds.copy()
            rez_flds.update(field_values)
            return rez_flds
        else:
            return field_values

    def action_result(
        self,
        result: Any = None,
        flds: Optional[Dict[str, Any]] = {},
        rez: Optional[BaseModel] = None,
        extra: Optional[Dict[str, Any]] = None,
        info: Optional[Dict[str, Any]] = None,
    ):
        return Rez(
            result=result,
            flds=flds,
            rez=rez,
            extra=extra,
            info=info if info else self.info(),
        )


class ActionRez(Action):
    action: Action
    rez: Rez


class RezDict(Action):
    rez: Rez
    dictionary: dict



from typing import List, Dict, Any, Optional
from enum import Enum
import json
import logging
from collections import namedtuple
from whendo.core.action import Action, Rez
from whendo.core.exception import TerminateSchedulerException


logger = logging.getLogger(__name__)


def processing_results(
    rez: Rez = None,
    processing_count: int = 0,
    success_count: int = 0,
    exception_count: int = 0,
    successful_actions: List[Dict[Any, Any]] = [],
    exception_actions: List[Dict[Any, Any]] = [],
):
    return {
        "result": rez,
        "processing_count": processing_count,
        "success_count": success_count,
        "exception_count": exception_count,
        "successful_actions": successful_actions,
        "exception_actions": exception_actions,
    }


class Vals(Action):
    """
    flds in this object override supplied rez.flds.
    """

    _vals: str = "_vals"
    vals: Optional[dict] = None

    def description(self):
        return f"This action adds ({self.vals}) to the field value flow."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        rez_flds = rez.flds.copy() if rez and rez.flds else {}
        rez_flds.update(flds.get("vals", {}))
        return self.action_result(
            result=rez.result if rez else None, rez=rez, flds=rez_flds
        )


class Result(Action):
    result: str = "result"
    value: Any

    def description(self):
        return f"This action returns the stored value as a Rez.result."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        value = flds["value"]
        return self.action_result(result=value, rez=rez, flds=rez.flds if rez else {})


class RaiseCmp(Action):
    """
    The execute method of this action will raise an exception if:

    1. both RaiseCmp.value, Rez.result are not None
    AND
    2. if RaiseCmp.cmp < 0 AND Rez.Result < RaiseCmp.value
       OR if RaiseCmp.cmp == 0 AND Rez.Result == RaiseCmp.value
       OR if RaiseCmp.cmp > 0 AND Rez.Result > RaiseCmp.value

    """

    raise_cmp: str = "raise_cmp"
    value: Optional[Any] = None
    cmp: Optional[int] = None

    def description(self):
        return f"This action raises an exception if the supplied result {self.operand_str()} the action's value"

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        result = rez.result if rez else None
        value = flds.get("value", None)
        if result is None or value is None:
            should_raise = False
            logger.info(
                f"one of RaiseCmp.value, Rez.result equals None; RaiseCmp ignored; info ({self.info()}"
            )
        else:
            cmp = flds.get("cmp", 0)
            should_raise = (
                result < value
                if cmp < 0
                else (result == value if cmp == 0 else result > value)
            )
        if should_raise:
            raise ValueError(
                f"exception thrown because ({value}) ({self.operand_str(flds['cmp'])}) ({result}); info ({self.info()})"
            )
        else:
            return self.action_result(
                result=result,
                rez=rez,
                flds=rez.flds if rez else {},
            )

    def operand_str(self, cmp: Optional[int] = None):
        if cmp is None:
            cmp = (
                self.cmp if self.cmp is not None else 0
            )  # compiler doesn't catch that self.cmp is not None
        return "<" if cmp < 0 else ("=" if cmp == 0 else ">")


class Terminate(Action):
    terminate: str = "terminate"

    def description(self):
        return f"This action raises a TerminateScheduler exception."

    def execute(self, tag: str = None, rez: Rez = None):
        raise TerminateSchedulerException(value=tag)


class Failure(Action):
    """ acts like False """

    failure: str = "failure"

    def description(self):
        return f"This action always fails. It serves a role similar to the bool, False."

    def execute(self, tag: str = None, rez: Rez = None):
        raise Exception("purposely unsuccessful execution", self.json())


class Success(Action):
    """ acts like True. It also serves as a fixed point execute implementation. """

    success: str = "success"

    def description(self):
        return (
            f"This action always succeeds. It serves a role similar to the bool, True."
        )

    def execute(self, tag: str = None, rez: Rez = None):
        return rez


class Fail(Action):
    """ acts like negation """

    fail: str = "fail"
    operand: Action

    def description(self):
        return f"If the operand ({self.operand}) fails, Not succeeds, otherwise fails (throwing an Exception). It serves a role similar to logical negation."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        operand = flds["operand"]
        operand_result = None
        try:
            operand_result = operand.execute(tag=tag, rez=rez)
        except Exception as exception:
            operand_result = exception
        if isinstance(operand_result, Exception):
            extra = {"execption": str(operand_result)}
            return self.action_result(
                result=True, rez=rez, flds=rez.flds if rez else {}, extra=extra
            )
        else:
            raise Exception(
                "exception generated; action execution treated as a failure",
                self.json(),
            )


class ListOpMode(str, Enum):
    ALL = "all"
    UNTIL_SUCCESS = "until_success"
    UNTIL_FAILURE = "until_failure"


class ListAction(Action):
    """
    executes actions based on:

        ListOpMode.ALL:   executes all, irrespective of individual outcomes
        ListOpMode.UNTIL_SUCCESS:    executes until the first successful action (no exception)
        ListOpMode.UNTIL_FAILURE:   executes until the first exception (failure)

    Intended to be abstract class; not intended to be instantiated. Its subclasses,
    All, Or, and And, should be used instead.
    """

    actions: List[Action] = []
    op_mode: Optional[ListOpMode] = None
    exception_on_no_success: Optional[bool] = None
    include_processing_info: Optional[bool] = None

    def description(self):
        if self.op_mode == ListOpMode.ALL:
            return (
                f"This action executes all of these actions in order: ({self.actions})."
            )
        if self.op_mode == ListOpMode.UNTIL_SUCCESS:
            return f"This action executes all of these actions in order until the first success: ({self.actions}). It serves a role similar to logical or."
        if self.op_mode == ListOpMode.UNTIL_FAILURE:
            return f"This action executes all of these actions in order until the first failure: ({self.actions}). It serves a role similar to logical and."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        processing_info = process_actions(
            rez=rez,
            op_mode=flds.get("op_mode", ListOpMode.ALL),
            tag=tag,
            actions=flds["actions"],
            successful_actions=[],
            exception_actions=[],
        )
        if (
            processing_info["processing_count"] > 0
            and processing_info["success_count"] == 0
            and flds.get("exception_on_no_success", False)
        ):
            raise Exception(
                "exception generated; all actions failing treated as a list action failure",
                json.dumps(self.dict()),
                json.dumps(processing_info),
            )
        extra = processing_info if flds.get("include_processing_info", False) else None
        rez = processing_info["result"]
        return self.action_result(
            result=rez.result if rez else None,
            rez=rez,
            flds=rez.flds if rez else {},
            extra=extra,
        )


class All(ListAction):
    """
    executes all actions
    """

    _all: str = "all"
    op_mode: ListOpMode = ListOpMode.ALL

    def execute(self, tag: str = None, rez: Rez = None):
        return super().execute(tag=tag, rez=rez)


class UntilSuccess(ListAction):
    """
    executes actions until first success
    """

    until_success: str = "until_success"
    op_mode: ListOpMode = ListOpMode.UNTIL_SUCCESS

    def execute(self, tag: str = None, rez: Rez = None):
        return super().execute(tag=tag, rez=rez)


class UntilFailure(ListAction):
    """
    executes actions until first failure
    """

    until_failure: str = "until_failure"
    op_mode: ListOpMode = ListOpMode.UNTIL_FAILURE

    def execute(self, tag: str = None, rez: Rez = None):
        return super().execute(tag=tag, rez=rez)


def process_actions(
    op_mode: ListOpMode,
    actions: List[Action],
    tag: str = None,
    rez: Rez = None,
    successful_actions: List[Dict[Any, Any]] = [],
    exception_actions: List[Dict[Any, Any]] = [],
    processing_count: int = 0,
    success_count: int = 0,
    exception_count: int = 0,
):
    """
    services the three action list classes, ALL, UNTIL_SUCCESS, and UNTIL_FAILURE
    """
    exceptions: List[Exception] = []
    loop_rez = rez
    for action in actions:
        exception = None
        try:
            loop_rez = action.execute(tag=tag, rez=loop_rez)
            logger.info(
                f"ListAction: tag ({tag}); executed action ({action}); loop data ({loop_rez})"
            )
        except Exception as e:
            exception = e
            logger.exception(
                f"ListAction: tag ({tag}); error while executing action ({action}); loop data ({loop_rez})",
                exc_info=exception,
            )
        processing_count += 1
        if exception:
            exception_count += 1
            exception_dict = action.dict().copy()
            exception_dict.update({"exception": str(exception)})
            exception_actions.append(exception_dict)
            if isinstance(exception, TerminateSchedulerException):
                exception.value = processing_results(
                    Rez(result=str(exception)),
                    processing_count,
                    success_count,
                    exception_count,
                    successful_actions,
                    exception_actions,
                )
                raise exception
            if op_mode == ListOpMode.UNTIL_FAILURE:  # stop after first failure
                break
            loop_rez = Rez(result=exception_dict, rez=rez, flds=rez.flds if rez else {})
        else:
            success_count += 1
            successful_actions.append(action.dict())
            if op_mode == ListOpMode.UNTIL_SUCCESS:  # stop after first success
                break
    return processing_results(
        loop_rez,
        processing_count,
        success_count,
        exception_count,
        successful_actions,
        exception_actions,
    )


class IfElse(Action):
    """
    executes actions based on:

        if test_action is successful, executes if_action
        otherwise executes else_action

    """

    if_else: str = "if_else"
    test_action: Action
    else_action: Action
    if_action: Optional[Action] = None
    exception_on_no_success: Optional[bool] = None
    include_processing_info: Optional[bool] = None

    def logger_info(self, s: str):
        logger.info(s)

    def logger_exception(self, s: str, exc_info: object):
        logger.info(s, exc_info)

    def description(self):
        return f"If  action ({self.test_action}) succeeds, then IfElse executes ({self.if_action}), otherwise executes ({self.else_action})"

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        test_action = flds["test_action"]
        else_action = flds["else_action"]
        if_action = flds.get("if_action", None)
        exception_on_no_success = flds.get("exception_on_no_success", False)
        include_processing_info = flds.get("include_processing_info", False)
        exception = None
        try:
            rez = test_action.execute(tag=tag, rez=rez)
            self.logger_info(
                f"IfElse: tag ({tag}); successfully executed 'test' action ({test_action})"
            )
        except Exception as e:
            exception = e
            self.logger_exception(
                f"IfElse: tag ({tag}); error while executing 'test' action ({test_action}) exception ({exception})",
                exc_info=exception,
            )
        processing_count = 1
        success_count = 0
        exception_count = 0
        successful_actions: List[Dict[Any, Any]] = []
        exception_actions: List[Dict[Any, Any]] = []
        if exception:  # execute the else action
            exception_count = 1
            exception_dict = test_action.dict().copy()
            exception_dict.update({"exception": str(exception)})
            exception_actions.append(exception_dict)
            if isinstance(exception, TerminateSchedulerException):
                exception.value = processing_results(
                    rez,
                    processing_count,
                    success_count,
                    exception_count,
                    successful_actions,
                    exception_actions,
                )
                raise exception
            exception = None
            try:
                rez = else_action.execute(tag=tag, rez=rez)
                self.logger_info(
                    f"IfElse: tag ({tag}); successfully executed 'else' action ({else_action})"
                )
            except Exception as e:
                exception = e
                self.logger_exception(
                    f"IfElse: tag ({tag}); error while executing 'else' action ({else_action}) exception ({exception})",
                    exc_info=exception,
                )
            processing_count += 1
            if exception:
                exception_count += 1
                exception_dict = else_action.dict().copy()
                exception_dict.update({"exception": str(exception)})
                exception_actions.append(exception_dict)
                if isinstance(exception, TerminateSchedulerException):
                    exception.value = processing_results(
                        rez,
                        processing_count,
                        success_count,
                        exception_count,
                        successful_actions,
                        exception_actions,
                    )
                    raise exception
                rez = self.action_result(
                    result=exception_dict, rez=rez, flds=rez.flds if rez else {}
                )
            else:
                success_count += 1
                successful_actions.append(else_action.dict())
        else:
            success_count += 1
            successful_actions.append(test_action.dict())

            if if_action:  # execute the if_action
                exception = None
                try:
                    rez = if_action.execute(tag=tag, rez=rez)
                    self.logger_info(
                        f"IfElse: tag ({tag}); successfully executed 'if' action ({if_action})"
                    )
                except Exception as e:
                    exception = e
                    self.logger_exception(
                        f"IfElse: tag ({tag}); error while executing 'if' action ({else_action}) exception ({exception})",
                        exc_info=exception,
                    )
                processing_count += 1
                if exception:
                    exception_count += 1
                    exception_dict = if_action.dict().copy()
                    exception_dict.update({"exception": str(exception)})
                    exception_actions.append(exception_dict)
                    if isinstance(exception, TerminateSchedulerException):
                        exception.value = processing_results(
                            rez,
                            processing_count,
                            success_count,
                            exception_count,
                            successful_actions,
                            exception_actions,
                        )
                        raise exception
                    rez = self.action_result(
                        result=exception_dict, rez=rez, flds=rez.flds if rez else {}
                    )
                else:
                    success_count += 1
                    successful_actions.append(if_action.dict())
            else:  # just use test_result since there is no if_action
                pass

        processing_info = processing_results(
            rez,
            processing_count,
            success_count,
            exception_count,
            successful_actions,
            exception_actions,
        )
        if (
            processing_info["processing_count"] > 0
            and processing_info["success_count"] == 0
            and exception_on_no_success
        ):
            raise Exception(
                "exception generated; all actions failing treated as an if-else action failure",
                json.dumps(self.dict()),
                json.dumps(processing_info),
            )
        extra = processing_info if include_processing_info else None
        rez = processing_info["result"]
        return self.action_result(
            result=rez.result if rez else None,
            rez=rez,
            flds=rez.flds if rez else {},
            extra=extra,
        )


class FormatOpMode(str, Enum):
    FLATTEN = "flatten"
    FIRST = "first"


class RezFmt(Action):
    rez_fmt: str = "rez_fmt"
    format_op_mode: Optional[FormatOpMode] = None

    def description(self):
        return f"Transforms Rez results recursively into a list structure or uses just the top-level rez."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        op_mode = flds.get("format_op_mode", FormatOpMode.FLATTEN)
        if rez is not None:
            if op_mode == FormatOpMode.FLATTEN:
                result = rez.flatten_results()
            else:
                result = rez.flatten_result()
        else:
            result = "Incoming Rez is empty."
        return self.action_result(
            result=result,
            rez=rez,
            flds=rez.flds if rez else {},
        )


# experimental


class Compose(Action):
    compose: str = "compose"
    actions: List[Action]

    def description(self):
        return (
            f"Performs function composition on supplied actions in left-to-right order."
        )

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        actions = flds["actions"]
        loop_rez = rez
        for action in actions:
            loop_rez = action.execute(tag=tag, rez=loop_rez)
        return loop_rez



from fastapi import APIRouter, status, Depends
from whendo.core.util import KeyTagMode
from whendo.api.shared import return_success, raised_exception, get_dispatcher
from whendo.core.resolver import resolve_server, resolve_rez, resolve_action
from whendo.core.action import RezDict

router = APIRouter(prefix="/servers", tags=["Servers"])


@router.get("", status_code=status.HTTP_200_OK)
def get_servers():
    try:
        return get_dispatcher(router).get_servers()
    except Exception as e:
        raise raised_exception(f"failed to retrieve servers", e)


@router.get("/{server_name}", status_code=status.HTTP_200_OK)
def get_server(server_name: str):
    try:
        server = get_dispatcher(router).get_server(server_name=server_name)
        return return_success(server)
    except Exception as e:
        raise raised_exception(f"failed to retrieve the server ({server_name})", e)


@router.post("/{server_name}", status_code=status.HTTP_200_OK)
def add_server(server_name: str, server=Depends(resolve_server)):
    try:
        assert server, f"couldn't resolve class for server ({server_name})"
        get_dispatcher(router).add_server(server_name=server_name, server=server)
        return return_success(f"server ({server_name}) was successfully added")
    except Exception as e:
        raise raised_exception(f"failed to add server ({server_name})", e)


@router.put("/{server_name}", status_code=status.HTTP_200_OK)
def set_server(server_name: str, server=Depends(resolve_server)):
    try:
        assert server, f"couldn't resolve class for server ({server_name})"
        get_dispatcher(router).set_server(server_name=server_name, server=server)
        return return_success(f"server ({server_name}) was successfully updated")
    except Exception as e:
        raise raised_exception(f"failed to update server ({server_name})", e)


@router.delete("/{server_name}", status_code=status.HTTP_200_OK)
def delete_server(server_name: str):
    try:
        get_dispatcher(router).delete_server(server_name=server_name)
        return return_success(f"server ({server_name}) was successfully deleted")
    except Exception as e:
        raise raised_exception(f"failed to delete server ({server_name})", e)


@router.get("/{server_name}/describe", status_code=status.HTTP_200_OK)
def describe_server(server_name: str):
    try:
        return get_dispatcher(router).describe_server(server_name=server_name)
    except Exception as e:
        raise raised_exception(f"failed to describe program ({server_name})", e)


@router.post("/{server_name}/add_key_tags", status_code=status.HTTP_200_OK)
def add_server_key_tags(server_name: str, key_tags: dict):
    try:
        get_dispatcher(router).add_server_key_tags(
            server_name=server_name, key_tags=key_tags
        )
        return return_success(
            f"key tags ({key_tags}) were successfully added to server ({server_name})."
        )
    except Exception as e:
        raise raised_exception(
            f"failed to add key-tags ({key_tags}) to server ({server_name}))", e
        )


@router.get("/{server_name}/get_tags", status_code=status.HTTP_200_OK)
def get_server_tags(server_name: str):
    try:
        return get_dispatcher(router).get_server_tags(server_name=server_name)
    except Exception as e:
        raise raised_exception(f"failed to retrieve tags for server ({server_name})", e)


@router.post("/by_tags/{mode}", status_code=status.HTTP_200_OK)
def get_servers_by_tags(mode: str, key_tags: dict):
    try:
        return get_dispatcher(router).get_servers_by_tags(
            key_tags=key_tags, key_tag_mode=KeyTagMode(mode)
        )
    except Exception as e:
        raise raised_exception(
            f"failed to get servers by key-tags ({key_tags}) using mode ({mode}))", e
        )


@router.get(
    "/{server_name}/actions/{action_name}/execute", status_code=status.HTTP_200_OK
)
def execute_on_server(server_name: str, action_name: str):
    try:
        return get_dispatcher(router).execute_on_server(
            server_name=server_name, action_name=action_name
        )
    except Exception as e:
        raise raised_exception(
            f"failed to get execute action ({action_name}) on server ({server_name})", e
        )


@router.post(
    "/{server_name}/actions/{action_name}/execute_with_rez",
    status_code=status.HTTP_200_OK,
)
def execute_on_server_with_rez(
    server_name: str, action_name: str, rez=Depends(resolve_rez)
):
    try:
        return get_dispatcher(router).execute_on_server_with_rez(
            server_name=server_name, action_name=action_name, rez=rez
        )
    except Exception as e:
        raise raised_exception(
            f"failed to execute action ({action_name}) on server ({server_name}) using rez ({rez})",
            e,
        )


@router.post(
    "/by_tags/{mode}/actions/{action_name}/execute", status_code=status.HTTP_200_OK
)
def execute_on_servers(action_name: str, mode: str, key_tags: dict):
    try:
        return get_dispatcher(router).execute_on_servers(
            action_name=action_name,
            key_tags=key_tags,
            key_tag_mode=KeyTagMode(mode),
        )
    except Exception as e:
        raise raised_exception(
            f"failed to execute action ({action_name}) by key tags ({key_tags})", e
        )


@router.post(
    "/by_tags/{mode}/actions/{action_name}/execute_with_rez",
    status_code=status.HTTP_200_OK,
)
def execute_on_servers_with_rez(
    action_name: str, mode: str, rez_dict=Depends(resolve_action)
):
    try:
        rez = rez_dict.rez
        key_tags = rez_dict.dictionary
        return get_dispatcher(router).execute_on_servers_with_rez(
            action_name=action_name,
            key_tags=key_tags,
            key_tag_mode=KeyTagMode(mode),
            rez=rez,
        )
    except Exception as e:
        raise raised_exception(
            f"failed to execute action ({action_name}) by rez-dict ({rez_dict})", e
        )



from datetime import datetime, timedelta, time
import os
import pathlib
from pydantic import BaseModel
from typing import List
import whendo.core.util as util
from whendo.core.util import Rez, DateTime, Now


def test_now_1():
    now1 = datetime.now()
    now2 = util.Now.dt()
    elapsed = now2 - now1
    assert elapsed < timedelta(microseconds=100.0)


def test_now_2():
    now = datetime.now()
    now2 = util.Now.t()
    date2 = datetime(
        year=now.year,
        month=now.month,
        day=now.day,
        hour=now2.hour,
        minute=now2.minute,
        second=now2.second,
        microsecond=now2.microsecond,
    )
    elapsed = date2 - now
    assert elapsed < timedelta(microseconds=100.0)


def test_now_3():
    now1 = datetime.now().time()
    now1_str = now1.strftime("%H:%M:%S")
    now2 = util.Now.s()
    assert now1_str in now2


def test_str_to_1():
    now = datetime.now()
    without_micro = datetime(
        year=now.year,
        month=now.month,
        day=now.day,
        hour=now.hour,
        minute=now.minute,
        second=now.second,
    )
    assert without_micro == util.str_to_dt(util.dt_to_str(now))


def test_str_to_2():
    now = datetime.now().time()
    without_micro = time(
        hour=now.hour,
        minute=now.minute,
        second=now.second,
    )
    assert without_micro == util.str_to_t(util.t_to_str(now))


def test_all_visible_subclasses():
    class A:
        pass

    class B(A):
        pass

    class C(A):
        pass

    class D(B):
        pass

    x = util.all_visible_subclasses(A)
    y = {B, C, D}
    assert x == y


def test_key_strings_from_class():
    class A(BaseModel):
        a: str
        b: str

    x = {"a", "b"}
    y = util.key_strings_from_class(A)
    assert x == y


def test_key_strings_from_dict():
    x = {"a", "b"}
    y = util.key_strings_from_dict({"a": 1, "b": 2})
    assert x == y


def test_resolve_instance():
    """
    Top > A, B, C
    C > D
    """

    class Top(BaseModel):
        pass

    class C(Top):
        c: str

    class D(C):
        d: str
        d_instance: Top

    class B(Top):
        b_instance: Top
        b_list: List[Top]

    class A(Top):
        a_list: List[Top]

    dictionary0 = {"c": "why"}
    dictionary1 = {  # D
        "c": "dc",
        "d": "abc",
        "d_instance": {  # C
            "c": "mmm",
        },
    }
    dictionary2 = {  # A
        "a_list": [
            {  # B
                "b_instance": {"c": "xxx"},  # C
                "b_list": [{"c": "yyy"}, {"c": "zzz"}, dictionary1],  # C  # C
            },
            {"c": "people", "d": "who", "d_instance": dictionary0},  # D
        ]
    }
    instance0 = util.resolve_instance(Top, dictionary0)
    instance1 = util.resolve_instance(Top, dictionary1)
    instance2 = util.resolve_instance(Top, dictionary2)
    assert isinstance(instance0, C)
    assert isinstance(instance1, D)
    assert isinstance(instance2, A)
    assert isinstance(instance2.a_list[0].b_instance, C)
    assert instance1 == instance2.a_list[0].b_list[2]
    assert instance0 == instance2.a_list[1].d_instance


def test_resolve_instance_multi_class_1():
    """
    Top > A, B, D
    C > D
    Top2 > W, X, Y
    Y > Z
    """

    class Top(BaseModel):
        pass

    class Top2(BaseModel):
        pass

    class C(Top):
        c: str

    class D(C):
        d: str
        d_instance: Top

    class B(Top):
        b_instance: Top2
        b_list: List[Top]

    class A(Top):
        a_list: List[Top]

    class Y(Top2):
        y: str

    class Z(Y):
        z: str
        z_instance: Top2

    class X(Top2):
        x_instance: Top2
        x_list: List[Top2]

    class W(Top2):
        w_list: List[Top2]

    dictionary0 = {"c": "why"}
    dictionary1 = {  # D
        "c": "dc",
        "d": "abc",
        "d_instance": {  # C
            "c": "mmm",
        },
    }
    dictionary2 = {  # A
        "a_list": [
            {  # B
                "b_instance": {"y": "bbb"},  # Y
                "b_list": [{"c": "yyy"}, {"c": "zzz"}, dictionary1],  # C  # C
            },
            {"c": "people", "d": "who", "d_instance": dictionary0},  # D
        ]
    }
    instance0 = util.resolve_instance_multi_class([Top, Top2], dictionary0)
    instance1 = util.resolve_instance_multi_class([Top, Top2], dictionary1)
    instance2 = util.resolve_instance_multi_class([Top, Top2], dictionary2)
    assert isinstance(instance0, C)
    assert isinstance(instance1, D)
    assert isinstance(instance2, A)
    assert isinstance(instance2.a_list[0].b_instance, Y)
    assert instance1 == instance2.a_list[0].b_list[2]
    assert instance0 == instance2.a_list[1].d_instance


def test_resolve_instance_multi_class_2():
    fp = util.FilePathe(path="/w/xy/z")
    dt = DateTime(dt=Now.dt())
    rezz = Rez(result=fp, flds={"x": dt})
    rez = Rez(result=dt, flds={"x": dt}, rez=rezz)
    rezd = rez.dict()
    rez2 = util.resolve_instance_multi_class([Rez, DateTime, util.FilePathe], rezd)
    assert isinstance(rez2.flds["x"], DateTime)
    assert rez2.result == dt
    assert rez2.rez == rezz
    assert isinstance(rez2.rez.flds["x"], DateTime)
    assert rez2.rez.result == fp


def test_filepathe():
    path = util.FilePathe(path="a/b/c")
    new_path = util.FilePathe(**(path.dict()))
    assert path.path == new_path.path, "path reconstruction failed"


def test_dirs(tmp_path):
    """
    test to make sure that files with the same name are written to
    separate directories
    """

    def assure_writes_and_reads(thunk):
        """ make sure that files can be written to and read from """
        lines = None
        file = thunk(home_path=tmp_path) + "foo.txt"
        with open(file, "w") as fid:
            fid.write("blee blee blee\n")
        with open(file, "r") as fid:
            lines = fid.readlines()
        assert lines is not None and isinstance(lines, list) and len(lines) >= 1
        return file

    file1 = assure_writes_and_reads(util.Dirs.saved_dir)
    file2 = assure_writes_and_reads(util.Dirs.output_dir)
    file3 = assure_writes_and_reads(util.Dirs.log_dir)
    assert file1 != file2 and file2 != file3 and file3 != file1


def test_dirs_2(tmp_path):
    try:
        file_path = os.path.join(util.Dirs.log_dir(home_path=tmp_path), "job.log")
    except:
        assert False


def test_output(tmp_path):
    file = str(tmp_path / "output.txt")
    util.Output.pprint({"what": "hmmm"}, file=file)
    lines = None
    with open(file, "r") as fid:
        lines = fid.readlines()
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1


def test_shared_rw_1():
    shared = util.SharedRWs.get("foo")

    def funk(dictionary: dict):
        dictionary["a"] = 1
        return 1

    result = shared.apply(funk)
    assert result == 1
    assert isinstance(shared.data_copy(), dict)


def test_shared_rw_2():
    """
    Test transactional nature with a funky failure.
    """
    shared = util.SharedRWs.get("foo")

    def funk1(dictionary: dict):
        dictionary["a"] = 1
        return 1

    result = shared.apply(funk1)

    def funk2(dictionary: dict):
        dictionary["a"] = 2
        raise Exception("oops")

    try:
        shared.apply(funk2)
    except:
        pass
    assert shared.data_copy()["a"] == 1


def test_shared_ro_1():
    dictionary = {"a": 1}
    shared = util.SharedROs.get("foo", dictionary)
    dictionary["a"] = 3
    assert shared.data_copy()["a"] == 1


def test_system_info():
    util.SystemInfo.init(host="127.0.0.4", port=8002)
    info = util.SystemInfo.get()
    assert "host" in info
    assert info["host"] == "127.0.0.4"
    assert "port" in info
    assert info["port"] == 8002
    assert "start" in info
    assert "cwd" in info
    assert "login" in info
    assert "os_version" in info
    assert "successes" in info
    assert "failures" in info
    assert "elapsed" in info
    assert "virtual_memory" in info



import time
import logging
from typing import Optional
from whendo.core.action import Action, Rez
import whendo.core.util as util


logger = logging.getLogger(__name__)


class SysInfo(Action):
    """
    Return system info
    """

    sys_info = "sys_info"

    def description(self):
        return f"This action returns system-level information."

    def execute(self, tag: str = None, rez: Rez = None):
        return self.action_result(
            result=util.SystemInfo.get(), rez=rez, flds=rez.flds if rez else {}
        )


class MiniInfo(Action):
    """
    Return terse local info.
    """

    mini_info = "mini_info"

    def description(self):
        return f"This action returns terse local information."

    def execute(self, tag: str = None, rez: Rez = None):
        return self.action_result(
            result=self.local_info(), rez=rez, flds=rez.flds if rez else {}
        )


class Pause(Action):
    """
    Sleep for supplied int seconds. It's best to make sure that the sleep duration
    does not interfere with job execution (blocking the thread).
    """

    seconds: Optional[float] = None
    pause: str = "pause"

    def description(self):
        return f"This action sleeps for {self.seconds} second{'s' if self.seconds != 1 else ''}."

    def execute(self, tag: str = None, rez: Rez = None):
        flds = self.compute_flds(rez=rez)
        seconds = flds.get("seconds", 1.0)
        time.sleep(seconds)
        return self.action_result(
            result=f"slept for ({seconds}) seconds",
            rez=rez,
            flds=rez.flds if rez else {},
        )
